"""TODO: Add a description here."""


import os

import spacy
from spacy.lang.en import English
from spacy.symbols import ORTH

import datasets

_CITATION_GENIA = """\
@article{article,
    author = {Kim, Jin-Dong and Ohta, Tomoko and Tateisi, Yuka and Tsujii, Jun'ichi},
    year = {2003},
    month = {02},
    pages = {i180-2},
    title = {GENIA corpus—A semantically annotated corpus for bio-textmining},
    volume = {19 Suppl 1},
    journal = {Bioinformatics (Oxford, England)},
    doi = {10.1093/bioinformatics/btg1023}
}"""

# You can copy an official description
_DESCRIPTION = """
The GENIA corpus is the primary collection of biomedical literature compiled and annotated within the scope
of the GENIA project. The corpus was created to support the development and evaluation of information
extraction and text mining systems for the domain of molecular biology.
"""

_HOMEPAGE = "http://www.geniaproject.org/genia-corpus/relation-corpus"

# TODO: Add the license for the dataset here if you can find it
_LICENSE = """\
GENIA Project License for Annotated Corpora

1. Copyright of abstracts

Any abstracts contained in this corpus are from PubMed(R), a database
of the U.S. National Library of Medicine (NLM).

NLM data are produced by a U.S. Government agency and include works of
the United States Government that are not protected by U.S. copyright
law but may be protected by non-US copyright law, as well as abstracts
originating from publications that may be protected by U.S. copyright
law.

NLM assumes no responsibility or liability associated with use of
copyrighted material, including transmitting, reproducing,
redistributing, or making commercial use of the data. NLM does not
provide legal advice regarding copyright, fair use, or other aspects
of intellectual property rights. Persons contemplating any type of
transmission or reproduction of copyrighted material such as abstracts
are advised to consult legal counsel.

2. Copyright of full texts

Any full texts contained in this corpus are from the PMC Open Access
Subset of PubMed Central (PMC), the U.S. National Institutes of Health
(NIH) free digital archive of biomedical and life sciences journal
literature.

Articles in the PMC Open Access Subset are protected by copyright, but
are made available under a Creative Commons or similar license that
generally allows more liberal redistribution and reuse than a
traditional copyrighted work. Please refer to the license of each
article for specific license terms.

3. Copyright of annotations

The copyrights of annotations created in the GENIA Project of Tsujii
Laboratory, University of Tokyo, belong in their entirety to the GENIA
Project.

4. Licence terms

Use and distribution of abstracts drawn from PubMed is subject to the
PubMed(R) license terms as stated in Clause 1.

Use and distribution of full texts is subject to the license terms
applying to each publication.

Annotations created by the GENIA Project are licensed under the
Creative Commons Attribution 3.0 Unported License. To view a copy of
this license, visit http://creativecommons.org/licenses/by/3.0/ or
send a letter to Creative Commons, 444 Castro Street, Suite 900,
Mountain View, California, 94041, USA.

Annotations created by the GENIA Project must be attributed as
detailed in Clause 5.

5. Attribution

The GENIA Project was founded and led by prof. Jun'ichi Tsujii and
the project and its annotation efforts have been coordinated in part
by Nigel Collier, Yuka Tateisi, Sang-Zoo Lee, Tomoko Ohta, Jin-Dong
Kim, and Sampo Pyysalo.

For a complete list of the GENIA Project members and contributors,
please refer to http://www.geniaproject.org.

The GENIA Project has been supported by Grant-in-Aid for Scientific
Research on Priority Area "Genome Information Science" (MEXT, Japan),
Grant-in-Aid for Scientific Research on Priority Area "Systems
Genomics" (MEXT, Japan), Core Research for Evolutional Science &
Technology (CREST) "Information Mobility Project" (JST, Japan),
Solution Oriented Research for Science and Technology (SORST) (JST,
Japan), Genome Network Project (MEXT, Japan) and Grant-in-Aid for
Specially Promoted Research (MEXT, Japan).

Annotations covered by this license must be attributed as follows:

    Corpus annotations (c) GENIA Project

Distributions including annotations covered by this licence must
include this license text and Attribution section.

6. References

- GENIA Project : http://www.geniaproject.org
- PubMed : http://www.pubmed.gov/
- NLM (United States National Library of Medicine) : http://www.nlm.nih.gov/
- MEXT (Ministry of Education, Culture, Sports, Science and Technology) : http://www.mext.go.jp/
- JST (Japan Science and Technology Agency) : http://www.jst.go.jp
"""

# TODO: Add link to the official dataset URLs here, currently test points to blind test file
# The HuggingFace dataset library don't host the datasets but only point to the original files
# This can be an arbitrary nested dict/list of URLs (see below in `_split_generators` method)
_DATA_URLs = {
    "train": "http://www.nactem.ac.uk/GENIA/current/GENIA-corpus/Relation/GENIA_relation_annotation_training_data.tar.gz",
    "dev": "http://www.nactem.ac.uk/GENIA/current/GENIA-corpus/Relation/GENIA_relation_annotation_development_data.tar.gz",
    # "test": "http://www.nactem.ac.uk/GENIA/current/GENIA-corpus/Relation/GENIA_relation_annotation_test_data.tar.gz"
}
# TODO: Add class labels
_CLASS_LABELS = ["Subunit-Complex", "Protein-Component"]


class Genia(datasets.GeneratorBasedBuilder):
    """TODO: Short description of my dataset."""

    VERSION = datasets.Version("1.0.0")  # type: ignore

    def _info(self):
        return datasets.DatasetInfo(
            # This is the description that will appear on the datasets page.
            description=_DESCRIPTION,
            # This defines the different columns of the dataset and their types
            features=datasets.Features(
                {
                    "tokens": datasets.Sequence(datasets.Value("string")),
                    "head_start": datasets.Value("int32"),
                    "head_end": datasets.Value("int32"),
                    "tail_start": datasets.Value("int32"),
                    "tail_end": datasets.Value("int32"),
                    "label": datasets.ClassLabel(names=_CLASS_LABELS),
                }
            ),
            # If there's a common (input, target) tuple from the features,
            # specify them here. They'll be used if as_supervised=True in
            # builder.as_dataset.
            supervised_keys=None,
            # Homepage of the dataset for documentation
            homepage=_HOMEPAGE,
            # License for the dataset if available
            license=_LICENSE,
            # Citation for the dataset
            citation=_CITATION_GENIA,
        )

    def _split_generators(self, dl_manager):
        """Returns SplitGenerators."""
        # TODO: This method is tasked with downloading/extracting the data and defining the splits depending on the configuration
        # If several configurations are possible (listed in BUILDER_CONFIGS), the configuration selected by the user is in self.config.name

        # dl_manager is a datasets.download.DownloadManager that can be used to download and extract URLs
        # It can accept any type or nested list/dict and will give back the same structure with the url replaced with path to local files.
        # By default the archives will be extracted and a path to a cached folder where they are extracted is returned instead of the archive
        data_files = dl_manager.download_and_extract(_DATA_URLs)

        return [
            datasets.SplitGenerator(
                name=datasets.Split.TRAIN,
                gen_kwargs={"filepath": data_files.get("train")},
            ),
            datasets.SplitGenerator(
                name=datasets.Split.VALIDATION,
                gen_kwargs={"filepath": data_files.get("dev")},
            ),
            # datasets.SplitGenerator(
            #     name=datasets.Split.TEST,
            #     gen_kwargs={"filepath": data_files.get("test")},
            # ),
        ]

    def _generate_examples(self, filepath):
        """Yields examples."""
        doc_ids, list_of_files = self._get_doc_ids_and_file_paths(filepath)
        processed_docs = self._get_processed_docs(doc_ids, list_of_files)

        idx = 0
        for doc in processed_docs:
            if "sentences" in doc and "sent_rels" in doc:
                sent_start_index = 0
                for sent, rels in zip(doc["sentences"], doc["sent_rels"]):
                    for rel in rels:
                        label = rel["label"]
                        head_start = rel["head_start"] - sent_start_index
                        head_end = rel["head_end"] - sent_start_index
                        tail_start = rel["tail_start"] - sent_start_index
                        tail_end = rel["tail_end"] - sent_start_index

                        id_ = str(idx) + "_" + doc["doc_id"]
                        idx += 1

                        yield id_, {
                            "tokens": sent["tokens"],
                            "head_start": head_start,
                            "head_end": head_end,
                            "tail_start": tail_start,
                            "tail_end": tail_end,
                            "label": label,
                        }

                    sent_start_index += len(sent)
            else:
                for rel in doc["relations"]:
                    label = rel["label"]
                    head_start = rel["head_start"]
                    head_end = rel["head_end"]
                    tail_start = rel["tail_start"]
                    tail_end = rel["tail_end"]

                    id_ = str(idx) + "_" + doc["doc_id"]
                    idx += 1

                    yield id_, {
                        "tokens": doc["tokens"],
                        "head_start": head_start,
                        "head_end": head_end,
                        "tail_start": tail_start,
                        "tail_end": tail_end,
                        "label": label,
                    }

    def _get_doc_ids_and_file_paths(self, path):
        list_of_files = {}
        for root, dirs, files in os.walk(path):
            for file in files:
                if file not in ["LICENSE", "README"]:
                    list_of_files[file] = os.path.join(root, file)
        doc_ids = list({file_name.split(".")[0] for file_name in list_of_files.keys()})
        doc_ids.sort()
        doc_ids.sort(key=len)
        return doc_ids, list_of_files

    def _get_processed_docs(self, doc_ids, list_of_files):
        ssplit = False
        try:
            nlp = spacy.load("en_core_web_sm")
            special_case = [{ORTH: "ca."}]
            nlp.tokenizer.add_special_case("ca.", special_case)
            ssplit = True
        except OSError as e:
            print(e)
            print(
                "You have to download the model first to enable sentence splitting: "
                "\tpython -m spacy download en_core_web_sm"
            )
            print("Resorting to tokenization only")
            nlp = English()
        processed_docs = []
        for doc_id in doc_ids:
            try:
                txt_file = list_of_files[doc_id + ".txt"]
                a1_file = list_of_files[doc_id + ".a1"]
                rel_file = list_of_files[doc_id + ".rel"]
            except KeyError:
                print(f"Missing annotation file for doc {doc_id}")
                continue

            relations = []
            entities = {}
            with open(txt_file, encoding="utf-8") as txt:
                text = txt.read()
            doc = nlp(text)
            with open(a1_file, encoding="utf-8") as a1:
                for line in a1.readlines():
                    if line.startswith("T"):
                        entity_id, entity = self._retrieve_entity(line, doc, doc_id)
                        entities[entity_id] = entity
            with open(rel_file, encoding="utf-8") as rel:
                for line in rel.readlines():
                    if line.startswith("T"):
                        entity_id, entity = self._retrieve_entity(line, doc, doc_id)
                        entities[entity_id] = entity
                    elif line.startswith("R"):
                        relations.append(self._retrieve_relation(line, entities))
            tokens = [token.text for token in doc]
            processed_doc = {
                "doc_id": doc_id,
                "text": text,
                "tokens": tokens,
                "entities": entities,
                "relations": relations,
            }
            if ssplit:
                sentences = self._convert_sentences(doc.sents)
                sentences = self._fix_ssplit(doc_id, sentences)
                sentence_tokens = []
                sentence_relations = []
                left_over_rels_indices = [True for _ in relations]
                for sent in sentences:
                    sent_rels = []
                    for idx, relation in enumerate(relations):
                        if (
                            min(relation["head_start"], relation["tail_start"]) >= sent["start"]
                            and max(relation["head_end"], relation["tail_end"]) <= sent["end"]
                        ):
                            sent_rels.append(relation)
                            left_over_rels_indices[idx] = False
                    sentence_tokens.append(sent["tokens"])
                    sentence_relations.append(sent_rels)
                left_over_rels = []
                for indicator, relation in zip(left_over_rels_indices, relations):
                    if indicator:
                        left_over_rels.append(relation)
                if left_over_rels:
                    print(
                        f"Examples in doc {doc_id} where spaCy ssplit were not compatible with relation annotation:"
                    )
                    print([list(sent) for sent in doc.sents])
                    print(sentences)
                    print(left_over_rels)
                processed_doc["sentences"] = sentences
                processed_doc["sent_rels"] = sentence_relations
            processed_docs.append(processed_doc)
        return processed_docs

    def _retrieve_entity(self, line, doc, doc_id=""):
        cols = line.strip().split()
        entity_id, _, start_char, end_char = cols[0:4]
        start_char, end_char = int(start_char), int(end_char)
        entity_type = " ".join(cols[4:])
        # default alignment mode is strict, but charOffset in annotation sometimes does not translate to token offsets
        # well, e.g. charOffsets only cover "LMP1" in "LMP1+"
        span = doc.char_span(start_char, end_char, alignment_mode="expand")
        if span:
            start, end = span.start, span.end
        else:
            snippet_start = max(0, start_char - 10)
            snippet_end = min(len(doc.text), end_char + 10)
            raise ValueError(
                f"{doc_id} Could not retrieve span for character offsets: "
                f"text[{start_char},{end_char}] = {doc.text[start_char:end_char]}\n"
                f"{doc.text[snippet_start:snippet_end]}\n"
                f"{list(doc)}"
            )
        return (entity_id, {"start": start, "end": end, "entity_type": entity_type})

    def _retrieve_relation(self, line, entities):
        cols = line.strip().split()
        relation_id, rel_type, arg1, arg2 = cols
        arg1 = arg1.split(":")[-1]
        head_start, head_end = entities[arg1]["start"], entities[arg1]["end"]
        arg2 = arg2.split(":")[-1]
        tail_start, tail_end = entities[arg2]["start"], entities[arg2]["end"]
        return {
            "rel_id": relation_id,
            "head_start": head_start,
            "head_end": head_end,
            "tail_start": tail_start,
            "tail_end": tail_end,
            "label": rel_type,
        }

    def _convert_sentences(self, sentences):
        sentence_dicts = []
        for sent in sentences:
            start, end = sent.start, sent.end
            tokens = [token.text for token in sent]
            sentence_dicts.append({"tokens": tokens, "start": start, "end": end})
        return sentence_dicts

    def _fix_ssplit(self, doc_id, sentences):
        if doc_id == "PMID-8164652":
            sentences[2]["tokens"] += sentences[3]["tokens"]
            sentences[2]["end"] = sentences[3]["end"]
            del sentences[3]
        elif doc_id == "PMID-9442380":
            sentences[4]["tokens"].append(sentences[5]["tokens"].pop(0))
            sentences[4]["end"] += 1
            sentences[5]["start"] += 1
        elif doc_id == "PMID-10201929":
            sentences[4]["tokens"] += sentences[5]["tokens"]
            sentences[4]["end"] = sentences[5]["end"]
            del sentences[5]
        elif doc_id == "PMID-10428853":
            sentences[3]["tokens"] += sentences[4]["tokens"]
            sentences[3]["end"] = sentences[4]["end"]
            del sentences[4]
        elif doc_id == "PMID-1675604":
            sentences[2]["tokens"] += sentences[3]["tokens"]
            sentences[2]["end"] = sentences[3]["end"]
            del sentences[3]
        sentences = [sent for sent in sentences if sent["tokens"]]
        return sentences



from .simple_transformer_text_classification import SimpleTransformerTextClassificationTaskModule
from .transformer_re_text_classification import TransformerRETextClassificationTaskModule
from .transformer_seq2seq import TransformerSeq2SeqTaskModule
from .transformer_span_classification import TransformerSpanClassificationTaskModule
from .transformer_text_classification import TransformerTextClassificationTaskModule
from .transformer_token_classification import TransformerTokenClassificationTaskModule



"""The SemEval-2010 Task 8 on Multi-way classification of semantic relations between pairs of nominals"""


import os
import re

import datasets

_CITATION = """\
@inproceedings{hendrickx-etal-2010-semeval,
    title = "{S}em{E}val-2010 Task 8: Multi-Way Classification of Semantic Relations between Pairs of Nominals",
    author = "Hendrickx, Iris  and
      Kim, Su Nam  and
      Kozareva, Zornitsa  and
      Nakov, Preslav  and
      {\'O} S{\'e}aghdha, Diarmuid  and
      Pad{\'o}, Sebastian  and
      Pennacchiotti, Marco  and
      Romano, Lorenza  and
      Szpakowicz, Stan",
    booktitle = "Proceedings of the 5th International Workshop on Semantic Evaluation",
    month = jul,
    year = "2010",
    address = "Uppsala, Sweden",
    publisher = "Association for Computational Linguistics",
    url = "https://www.aclweb.org/anthology/S10-1006",
    pages = "33--38",
}
"""

_DESCRIPTION = """\
The SemEval-2010 Task 8 focuses on Multi-way classification of semantic relations between pairs of nominals.
The task was designed to compare different approaches to semantic relation classification
and to provide a standard testbed for future research.
"""

_URL = "https://drive.google.com/uc?export=download&id=0B_jQiLugGTAkMDQ5ZjZiMTUtMzQ1Yy00YWNmLWJlZDYtOWY1ZDMwY2U4YjFk"

_CLASS_LABELS = [
    "Cause-Effect(e1,e2)",
    "Cause-Effect(e2,e1)",
    "Component-Whole(e1,e2)",
    "Component-Whole(e2,e1)",
    "Content-Container(e1,e2)",
    "Content-Container(e2,e1)",
    "Entity-Destination(e1,e2)",
    "Entity-Destination(e2,e1)",
    "Entity-Origin(e1,e2)",
    "Entity-Origin(e2,e1)",
    "Instrument-Agency(e1,e2)",
    "Instrument-Agency(e2,e1)",
    "Member-Collection(e1,e2)",
    "Member-Collection(e2,e1)",
    "Message-Topic(e1,e2)",
    "Message-Topic(e2,e1)",
    "Product-Producer(e1,e2)",
    "Product-Producer(e2,e1)",
    "Other",
]


class SemEval2010Task8(datasets.GeneratorBasedBuilder):
    """The SemEval-2010 Task 8 focuses on Multi-way classification of semantic relations between pairs of nominals.
    The task was designed to compare different approaches to semantic relation classification
    and to provide a standard testbed for future research."""

    VERSION = datasets.Version("1.0.0")  # type: ignore

    def _info(self):
        return datasets.DatasetInfo(
            # This is the description that will appear on the datasets page.
            description=_DESCRIPTION,
            # This defines the different columns of the dataset and their types
            features=datasets.Features(
                {
                    "tokens": datasets.Sequence(datasets.Value("string")),
                    "head_start": datasets.Value("int32"),
                    "head_end": datasets.Value("int32"),
                    "tail_start": datasets.Value("int32"),
                    "tail_end": datasets.Value("int32"),
                    "label": datasets.ClassLabel(names=_CLASS_LABELS),
                }
            ),
            # If there's a common (input, target) tuple from the features,
            # specify them here. They'll be used if as_supervised=True in
            # builder.as_dataset.
            supervised_keys=None,
            # Homepage of the dataset for documentation
            homepage="https://semeval2.fbk.eu/semeval2.php?location=tasks&taskid=11",
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager):
        """Returns SplitGenerators."""
        # dl_manager is a datasets.download.DownloadManager that can be used to
        # download and extract URLs
        dl_dir = dl_manager.download_and_extract(_URL)
        data_dir = os.path.join(dl_dir, "SemEval2010_task8_all_data")
        return [
            datasets.SplitGenerator(
                name=datasets.Split.TRAIN,
                # These kwargs will be passed to _generate_examples
                gen_kwargs={
                    "filepath": os.path.join(
                        data_dir, "SemEval2010_task8_training/TRAIN_FILE.TXT"
                    ),
                },
            ),
            datasets.SplitGenerator(
                name=datasets.Split.TEST,
                gen_kwargs={
                    "filepath": os.path.join(
                        data_dir, "SemEval2010_task8_testing_keys/TEST_FILE_FULL.TXT"
                    ),
                },
            ),
        ]

    def _generate_examples(self, filepath):
        """Yields examples."""
        with open(filepath, encoding="utf-8") as f:
            raw_lines = []
            for line in f:
                line = line.strip()

                if not line:
                    idx, example = self._raw_lines_to_example(raw_lines)
                    yield idx, example
                    raw_lines = []
                    continue

                raw_lines.append(line)

    def _raw_lines_to_example(self, raw_lines):
        raw_id, raw_text = raw_lines[0].split("\t")
        label = raw_lines[1]
        id_ = int(raw_id)
        raw_text = raw_text.strip('"')

        # Some special cases (e.g., missing spaces before entity marker)
        if id_ in [213, 4612, 6373, 8411, 9867]:
            raw_text = raw_text.replace("<e2>", " <e2>")
        if id_ in [2740, 4219, 4784]:
            raw_text = raw_text.replace("<e1>", " <e1>")
        if id_ == 9256:
            raw_text = raw_text.replace("log- jam", "log-jam")

        # necessary if text should be whitespace tokenizeable
        if id_ in [2609, 7589]:
            raw_text = raw_text.replace("1 1/2", "1-1/2")
        if id_ == 10591:
            raw_text = raw_text.replace("1 1/4", "1-1/4")
        if id_ == 10665:
            raw_text = raw_text.replace("6 1/2", "6-1/2")

        raw_text = re.sub(r"([.,!?()])$", r" \1", raw_text)
        raw_text = re.sub(r"(e[12]>)([',;:\"\(\)])", r"\1 \2", raw_text)
        raw_text = re.sub(r"([',;:\"\(\)])(</?e[12])", r"\1 \2", raw_text)
        raw_text = raw_text.replace("<e1>", "<e1> ")
        raw_text = raw_text.replace("<e2>", "<e2> ")
        raw_text = raw_text.replace("</e1>", " </e1>")
        raw_text = raw_text.replace("</e2>", " </e2>")

        tokens = raw_text.split(" ")

        head_start = tokens.index("<e1>")
        tokens.pop(head_start)

        head_end = tokens.index("</e1>")
        tokens.pop(head_end)

        tail_start = tokens.index("<e2>")
        tokens.pop(tail_start)

        tail_end = tokens.index("</e2>")
        tokens.pop(tail_end)

        return id_, {
            "tokens": tokens,
            "head_start": head_start,
            "head_end": head_end,
            "tail_start": tail_start,
            "tail_end": tail_end,
            "label": label,
        }



"""TODO: Add a description here."""


import json

import datasets

# TODO: Add the licence for the dataset here if you can find it
_LICENSE = ""

# TODO: Add link to the official dataset URLs here
# The HuggingFace dataset library don't host the datasets but only point to the original files
# This can be an arbitrary nested dict/list of URLs (see below in `_split_generators` method)

_CITATION_FEWREL_1 = """\
@inproceedings{han-etal-2018-fewrel,
    title = "{F}ew{R}el: A Large-Scale Supervised Few-Shot Relation Classification Dataset with State-of-the-Art Evaluation",
    author = "Han, Xu and Zhu, Hao and Yu, Pengfei and Wang, Ziyun and Yao, Yuan and Liu, Zhiyuan and Sun, Maosong",
    booktitle = "Proceedings of the 2018 Conference on Empirical Methods in Natural Language Processing",
    month = oct # "-" # nov,
    year = "2018",
    address = "Brussels, Belgium",
    publisher = "Association for Computational Linguistics",
    url = "https://www.aclweb.org/anthology/D18-1514",
    doi = "10.18653/v1/D18-1514",
    pages = "4803--4809"
}"""

_CITATION_FEWREL_2 = """\
@inproceedings{han-etal-2018-fewrel,
    title = "{F}ew{R}el: A Large-Scale Supervised Few-Shot Relation Classification Dataset with State-of-the-Art Evaluation",
    author = "Han, Xu and Zhu, Hao and Yu, Pengfei and Wang, Ziyun and Yao, Yuan and Liu, Zhiyuan and Sun, Maosong",
    booktitle = "Proceedings of the 2018 Conference on Empirical Methods in Natural Language Processing",
    month = oct # "-" # nov,
    year = "2018",
    address = "Brussels, Belgium",
    publisher = "Association for Computational Linguistics",
    url = "https://www.aclweb.org/anthology/D18-1514",
    doi = "10.18653/v1/D18-1514",
    pages = "4803--4809"
}

@inproceedings{gao-etal-2019-fewrel,
    title = "{F}ew{R}el 2.0: Towards More Challenging Few-Shot Relation Classification",
    author = "Gao, Tianyu and Han, Xu and Zhu, Hao and Liu, Zhiyuan and Li, Peng and Sun, Maosong and Zhou, Jie",
    booktitle = "Proceedings of the 2019 Conference on Empirical Methods in Natural Language Processing and the 9th International Joint Conference on Natural Language Processing (EMNLP-IJCNLP)",
    month = nov,
    year = "2019",
    address = "Hong Kong, China",
    publisher = "Association for Computational Linguistics",
    url = "https://www.aclweb.org/anthology/D19-1649",
    doi = "10.18653/v1/D19-1649",
    pages = "6251--6256"
}
"""


class FewRelConfig(datasets.BuilderConfig):
    """BuilderConfig for FewRel."""

    def __init__(
        self,
        data_url,
        citation,
        url,
        class_labels,
        description,
        **kwargs,
    ):
        """BuilderConfig for FewRel.
        Args:
          data_url: `string`, url to download the zip file from
          citation: `string`, citation for the data set
          url: `string`, url for information about the data set
          class_labels: `list[string]`, the list of classes if the label is
            categorical. If not provided, then the label will be of type
            `datasets.Value('float32')`.
          **kwargs: keyword arguments forwarded to super.
        """
        super().__init__(version=datasets.Version("1.0.0", ""), **kwargs)
        self.class_labels = class_labels
        self.data_url = data_url
        self.citation = citation
        self.url = url
        self.description = description


# TODO: Name of the dataset usually match the script name with CamelCase instead of snake_case
class FewRel(datasets.GeneratorBasedBuilder):
    """TODO: Short description of my dataset."""

    BUILDER_CONFIGS = [
        FewRelConfig(
            name="fewrel_train",
            data_url="https://github.com/thunlp/FewRel/raw/master/data/train_wiki.json",
            citation=_CITATION_FEWREL_1,
            url="https://thunlp.github.io/1/fewrel1.html",
            class_labels=[
                "P931",
                "P4552",
                "P140",
                "P1923",
                "P150",
                "P6",
                "P27",
                "P449",
                "P1435",
                "P175",
                "P1344",
                "P39",
                "P527",
                "P740",
                "P706",
                "P84",
                "P495",
                "P123",
                "P57",
                "P22",
                "P178",
                "P241",
                "P403",
                "P1411",
                "P135",
                "P991",
                "P156",
                "P176",
                "P31",
                "P1877",
                "P102",
                "P1408",
                "P159",
                "P3373",
                "P1303",
                "P17",
                "P106",
                "P551",
                "P937",
                "P355",
                "P710",
                "P137",
                "P674",
                "P466",
                "P136",
                "P306",
                "P127",
                "P400",
                "P974",
                "P1346",
                "P460",
                "P86",
                "P118",
                "P264",
                "P750",
                "P58",
                "P3450",
                "P105",
                "P276",
                "P101",
                "P407",
                "P1001",
                "P800",
                "P131",
            ],
            description="",
        ),
        FewRelConfig(
            name="fewrel_validation",
            data_url="https://github.com/thunlp/FewRel/raw/master/data/val_wiki.json",
            citation=_CITATION_FEWREL_1,
            url="https://thunlp.github.io/1/fewrel1.html",
            class_labels=[
                "P177",
                "P364",
                "P2094",
                "P361",
                "P641",
                "P59",
                "P413",
                "P206",
                "P412",
                "P155",
                "P26",
                "P410",
                "P25",
                "P463",
                "P40",
                "P921",
            ],
            description="",
        ),
        FewRelConfig(
            name="fewrel2_validation",
            data_url="https://github.com/thunlp/FewRel/raw/master/data/val_pubmed.json",
            citation=_CITATION_FEWREL_2,
            url="https://thunlp.github.io/2/fewrel2_da.html",
            class_labels=[
                "biological_process_involves_gene_product",
                "inheritance_type_of",
                "is_normal_tissue_origin_of_disease",
                "ingredient_of",
                "is_primary_anatomic_site_of_disease",
                "gene_found_in_organism",
                "occurs_in",
                "causative_agent_of",
                "classified_as",
                "gene_plays_role_in_process",
            ],
            description="",
        ),
    ]

    def _info(self):
        features = datasets.Features(
            {
                "tokens": datasets.Sequence(datasets.Value("string")),
                "head_start": datasets.Value("int32"),
                "head_end": datasets.Value("int32"),
                "tail_start": datasets.Value("int32"),
                "tail_end": datasets.Value("int32"),
                "label": datasets.ClassLabel(names=self.config.class_labels),
            }
        )

        return datasets.DatasetInfo(
            # This is the description that will appear on the datasets page.
            description=self.config.description,
            # This defines the different columns of the dataset and their types
            features=features,  # Here we define them above because they are different between the two configurations
            # If there's a common (input, target) tuple from the features,
            # specify them here. They'll be used if as_supervised=True in
            # builder.as_dataset.
            supervised_keys=None,
            # Homepage of the dataset for documentation
            homepage=self.config.url,
            # License for the dataset if available
            license=_LICENSE,
            # Citation for the dataset
            citation=self.config.citation,
        )

    def _split_generators(self, dl_manager):
        """Returns SplitGenerators."""
        # TODO: This method is tasked with downloading/extracting the data and defining the splits depending on the configuration
        # If several configurations are possible (listed in BUILDER_CONFIGS), the configuration selected by the user is in self.config.name

        # dl_manager is a datasets.download.DownloadManager that can be used to download and extract URLs
        # It can accept any type or nested list/dict and will give back the same structure with the url replaced with path to local files.
        # By default the archives will be extracted and a path to a cached folder where they are extracted is returned instead of the archive
        filepath = dl_manager.download_and_extract(self.config.data_url)

        split = (
            datasets.Split.VALIDATION if "validation" in self.config.name else datasets.Split.TRAIN
        )

        return [
            datasets.SplitGenerator(
                name=split,
                gen_kwargs={"filepath": filepath},
            ),
        ]

    def _generate_examples(self, filepath):
        """Yields examples."""
        with open(filepath, encoding="utf-8") as f:
            data = json.load(f)
            for label, examples in data.items():
                for idx, example in enumerate(examples):
                    id_ = label + "_" + str(idx)

                    head_token_positions = example["h"][2][0]
                    tail_token_positions = example["t"][2][0]

                    head_start = head_token_positions[0]
                    head_end = head_token_positions[-1]
                    tail_start = tail_token_positions[0]
                    tail_end = tail_token_positions[-1]

                    yield id_, {
                        "tokens": example["tokens"],
                        "head_start": head_start,
                        "head_end": head_end + 1,  # make end offset exclusive
                        "tail_start": tail_start,
                        "tail_end": tail_end + 1,  # make end offset exclusive
                        "label": label,
                    }



from pytorch_ie.models.transformer_seq2seq import (
    TransformerSeq2SeqModel,
    TransformerSeq2SeqModelBatchEncoding,
    TransformerSeq2SeqModelBatchOutput,
    TransformerSeq2SeqModelStepBatchEncoding,
)
from pytorch_ie.models.transformer_span_classification import (
    TransformerSpanClassificationModel,
    TransformerSpanClassificationModelBatchEncoding,
    TransformerSpanClassificationModelBatchOutput,
    TransformerSpanClassificationModelStepBatchEncoding,
)
from pytorch_ie.models.transformer_text_classification import (
    TransformerTextClassificationModel,
    TransformerTextClassificationModelBatchEncoding,
    TransformerTextClassificationModelBatchOutput,
    TransformerTextClassificationModelStepBatchEncoding,
)
from pytorch_ie.models.transformer_token_classification import (
    TransformerTokenClassificationModel,
    TransformerTokenClassificationModelBatchEncoding,
    TransformerTokenClassificationModelBatchOutput,
    TransformerTokenClassificationModelStepBatchEncoding,
)

__all__ = [
    "TransformerSeq2SeqModel",
    "TransformerSeq2SeqModelBatchEncoding",
    "TransformerSeq2SeqModelBatchOutput",
    "TransformerSeq2SeqModelStepBatchEncoding",
    "TransformerSpanClassificationModel",
    "TransformerSpanClassificationModelBatchEncoding",
    "TransformerSpanClassificationModelBatchOutput",
    "TransformerSpanClassificationModelStepBatchEncoding",
    "TransformerTextClassificationModel",
    "TransformerTextClassificationModelBatchEncoding",
    "TransformerTextClassificationModelBatchOutput",
    "TransformerTextClassificationModelStepBatchEncoding",
    "TransformerTokenClassificationModel",
    "TransformerTokenClassificationModelBatchEncoding",
    "TransformerTokenClassificationModelBatchOutput",
    "TransformerTokenClassificationModelStepBatchEncoding",
]



from functools import wraps
from inspect import Signature, isclass, signature
from typing import Callable, Dict, Iterable, List, Optional, Set, Tuple, Type, TypeVar, Union

import pandas as pd
from datasets.formatting import _register_formatter

import datasets
from pytorch_ie.core.document import Document
from pytorch_ie.data.dataset_formatter import DocumentFormatter

_register_formatter(DocumentFormatter, "document")


def decorate_convert_to_dict_of_lists(f):
    """
    Decorate the mapped function, so that converts a single Document to a dict,
    and a list of Documents into a dict of lists.
    """

    @wraps(f)
    def decorated(item, *args, **kwargs):
        if isinstance(item, list):
            # Convert a list of dicts into a dict of lists.
            return pd.DataFrame([e.asdict() for e in f(item, *args, **kwargs)]).to_dict(
                orient="list"
            )
        else:
            return f(item, *args, **kwargs).asdict()

    return decorated


E = TypeVar("E")


def dl_to_ld(dict_list: Dict[str, List[E]]) -> List[Dict[str, E]]:
    # Convert a dict of lists to a list of dicts
    return [dict(zip(dict_list, t)) for t in zip(*dict_list.values())]


def ld_to_dl(
    list_dict: List[Dict[str, E]], keys: Optional[Iterable[str]] = None
) -> Dict[str, List[E]]:
    # Convert a list of dicts to a dict of lists.
    # Provide keys to create the expected format when lists are empty.
    if keys is None:
        keys = list_dict[0]
    return {k: [dic[k] for dic in list_dict] for k in keys}


def decorate_convert_to_document_and_back(f, document_type: Type[Document], batched: bool):
    @wraps(f)
    def decorated(item, *args, **kwargs):
        if batched:
            # Convert a list of dicts into a dict of lists.
            return ld_to_dl(
                [
                    e.asdict()
                    for e in f(
                        [document_type.fromdict(x) for x in dl_to_ld(item)], *args, **kwargs
                    )
                ],
                # passing the keys allows to work correctly with empty lists
                keys=item.keys(),
            )
        else:
            return f(document_type.fromdict(item), *args, **kwargs).asdict()

    return decorated


def _check_fields_for_casting(
    field_mapping: Dict[str, str],
    current_document_type: Type[Document],
    new_document_type: Type[Document],
    column_names: list[str],
) -> Tuple[Set[str], Set[str]]:
    original_fields = {field.name: field for field in current_document_type.fields()}
    new_fields = {field.name: field for field in new_document_type.fields()}
    hidden_fields = set(column_names) - set(original_fields)
    fields_to_map_not_in_original_fields = (
        set(field_mapping) - set(original_fields) - set(hidden_fields)
    )
    if len(fields_to_map_not_in_original_fields) > 0:
        raise ValueError(
            f"some fields to rename are not in the original document_type or hidden fields: "
            f"{fields_to_map_not_in_original_fields}"
        )
    mapped_but_not_in_new_fields = set(field_mapping.values()) - set(new_fields)
    if len(mapped_but_not_in_new_fields) > 0:
        raise ValueError(
            f"some renamed fields are not in the new document_type: {mapped_but_not_in_new_fields}"
        )
    original_fields_mapped = {
        field_mapping.get(f_name, f_name): f for f_name, f in original_fields.items()
    }
    added_field_names = set(new_fields) - set(original_fields_mapped)
    removed_field_names = set(original_fields) - set(new_fields) - set(field_mapping)

    # Sanity checks
    kept_field_names = set(original_fields_mapped) & set(new_fields)
    for f_name_mapped in kept_field_names:
        f = original_fields_mapped[f_name_mapped]
        new_f = new_fields[f_name_mapped]
        if not (
            f.type == new_f.type
            and f.default == new_f.default
            and f.default_factory == new_f.default_factory
        ):
            raise ValueError(f"new field is not the same as old field:\n{new_f}\nvs\n{f}")

    return removed_field_names, added_field_names


D = TypeVar("D", bound=Document)


def _infer_document_type_from_function_return(function: Callable) -> Optional[Type[Document]]:
    # try to infer the document type from the return type annotation of function
    return_signature = signature(function).return_annotation
    if not return_signature == Signature.empty:
        if not isclass(return_signature) or not issubclass(return_signature, Document):
            raise TypeError(
                f"the return type annotation of the function used with map is not a subclass of Document"
            )
        return return_signature
    return None


class Dataset(datasets.Dataset):
    def __init__(
        self,
        document_type: Type[Document],
        arrow_table: datasets.table.Table,
        info: Optional[datasets.DatasetInfo] = None,
        split: Optional[datasets.NamedSplit] = None,
        indices_table: Optional[datasets.table.Table] = None,
        fingerprint: Optional[str] = None,
    ):
        super().__init__(
            arrow_table=arrow_table,
            info=info,
            split=split,
            indices_table=indices_table,
            fingerprint=fingerprint,
        )

        self.document_type = document_type
        self.set_format("document", document_type=document_type)

    @classmethod
    def get_base_kwargs(cls, dataset: datasets.Dataset):
        return dict(
            arrow_table=dataset._data,
            info=dataset.info,
            split=dataset.split,
            indices_table=dataset._indices,
            fingerprint=dataset._fingerprint,
        )

    @classmethod
    def from_hf_dataset(
        cls, dataset: datasets.Dataset, document_type: Type[Document]
    ) -> "Dataset":
        document_dataset = cls(document_type=document_type, **cls.get_base_kwargs(dataset))
        return document_dataset

    def apply_hf_func(self, func, **kwargs) -> "Dataset":
        return Dataset.from_hf_dataset(
            func(self, **kwargs),
            document_type=self.document_type,
        )

    def map(
        self,
        function: Optional[Callable] = None,
        with_indices: bool = False,
        with_rank: bool = False,
        input_columns: Optional[Union[str, List[str]]] = None,
        batched: bool = False,
        batch_size: Optional[int] = 1000,
        drop_last_batch: bool = False,
        remove_columns: Optional[Union[str, List[str]]] = None,
        keep_in_memory: bool = False,
        load_from_cache_file: Optional[bool] = None,
        cache_file_name: Optional[str] = None,
        writer_batch_size: Optional[int] = 1000,
        features: Optional[datasets.Features] = None,
        disable_nullable: bool = False,
        fn_kwargs: Optional[dict] = None,
        num_proc: Optional[int] = None,
        suffix_template: str = "_{rank:05d}_of_{num_proc:05d}",
        new_fingerprint: Optional[str] = None,
        desc: Optional[str] = None,
        as_documents: bool = True,
        result_document_type: Optional[Type[Document]] = None,
    ) -> "Dataset":
        dataset = super().map(
            function=decorate_convert_to_dict_of_lists(function) if as_documents else function,
            with_indices=with_indices,
            with_rank=with_rank,
            input_columns=input_columns,
            batched=batched,
            batch_size=batch_size,
            drop_last_batch=drop_last_batch,
            remove_columns=remove_columns,
            keep_in_memory=keep_in_memory,
            # ignore typing because typing in Huggingface Dataset.map() is incorrect
            load_from_cache_file=load_from_cache_file,  # type: ignore
            cache_file_name=cache_file_name,
            writer_batch_size=writer_batch_size,
            features=features,
            disable_nullable=disable_nullable,
            fn_kwargs=fn_kwargs,
            num_proc=num_proc,
            suffix_template=suffix_template,
            new_fingerprint=new_fingerprint,
            desc=desc,
        )

        if result_document_type is None:
            if function is not None and as_documents:
                result_document_type = _infer_document_type_from_function_return(function=function)
            if result_document_type is None:
                result_document_type = self.document_type

        return Dataset.from_hf_dataset(dataset, document_type=result_document_type)

    def cast_document_type(
        self,
        new_document_type: Type[D],
        remove_columns: bool = False,
        field_mapping: Optional[Dict[str, str]] = None,
    ) -> "Dataset":
        field_mapping = field_mapping or {}

        removed_field_names, added_field_names = _check_fields_for_casting(
            field_mapping=field_mapping,
            current_document_type=self.document_type,
            new_document_type=new_document_type,
            column_names=self.column_names,
        )

        new_hf_dataset = datasets.Dataset(**self.get_base_kwargs(self))

        if remove_columns:
            new_hf_dataset = new_hf_dataset.remove_columns(list(removed_field_names))

        rename_targets_already_in_columns = (
            set(field_mapping.values()) - set(field_mapping)
        ) & set(new_hf_dataset.column_names)
        if len(rename_targets_already_in_columns) > 0:
            raise ValueError(
                f"rename targets are already in column names: {rename_targets_already_in_columns}. Did you miss "
                f"to set remove_columns=True in a previous call of cast_document_type?"
            )

        new_hf_dataset = new_hf_dataset.rename_columns(field_mapping)
        for f_name in added_field_names:
            if f_name not in new_hf_dataset.column_names:
                # add empty columns
                new_hf_dataset = new_hf_dataset.add_column(
                    name=f_name, column=len(new_hf_dataset) * [{}]
                )
        new_dataset = Dataset.from_hf_dataset(new_hf_dataset, document_type=new_document_type)

        return new_dataset


class IterableDataset(datasets.IterableDataset):
    def __init__(
        self, document_type: Type[Document], hidden_columns: Optional[Set[str]] = None, **kwargs
    ):
        super().__init__(**kwargs)
        self.document_type = document_type
        self._document_field_names = [field.name for field in document_type.fields()]
        self.hidden_columns = set()
        if hidden_columns is not None:
            self.hidden_columns.update(hidden_columns)

    @property
    def column_names(self) -> List[str]:
        return self._document_field_names + list(self.hidden_columns)

    @classmethod
    def get_base_kwargs(cls, dataset: datasets.IterableDataset):
        return dict(
            ex_iterable=dataset._ex_iterable,
            info=dataset.info,
            split=dataset.split,
            format_type=dataset._format_type,
            shuffling=dataset._shuffling,
            token_per_repo_id=dataset._token_per_repo_id,
        )

    @classmethod
    def from_hf_dataset(
        cls,
        dataset: datasets.IterableDataset,
        document_type: Type[Document],
        hidden_columns: Optional[Set[str]] = None,
    ) -> "IterableDataset":
        dataset = cls(
            document_type=document_type,
            hidden_columns=hidden_columns,
            **cls.get_base_kwargs(dataset),
        )
        return dataset

    def __iter__(self):
        for example in iter(super().__iter__()):
            yield self.document_type.fromdict(example)

    def map(  # type: ignore
        self,
        function: Optional[Callable] = None,
        batched: bool = False,
        as_documents: bool = True,
        result_document_type: Optional[Type[Document]] = None,
        **kwargs,
    ) -> "IterableDataset":
        dataset_mapped = super().map(
            function=decorate_convert_to_document_and_back(
                function, document_type=self.document_type, batched=batched
            )
            if as_documents
            else function,
            batched=batched,
            **kwargs,
        )

        if result_document_type is None:
            if function is not None and as_documents:
                result_document_type = _infer_document_type_from_function_return(function=function)
            if result_document_type is None:
                result_document_type = self.document_type

        return IterableDataset.from_hf_dataset(dataset_mapped, document_type=result_document_type)

    def apply_hf_func(self, func, **kwargs) -> "IterableDataset":
        return IterableDataset.from_hf_dataset(
            func(self, **kwargs),
            document_type=self.document_type,
            hidden_columns=self.hidden_columns,
        )

    def cast_document_type(
        self,
        new_document_type: Type[D],
        remove_columns: bool = False,
        field_mapping: Optional[Dict[str, str]] = None,
    ) -> "IterableDataset":
        field_mapping = field_mapping or {}

        removed_field_names, added_field_names = _check_fields_for_casting(
            field_mapping=field_mapping,
            current_document_type=self.document_type,
            new_document_type=new_document_type,
            column_names=self.column_names,
        )
        hidden_columns = set(self.hidden_columns)
        new_hf_dataset = datasets.IterableDataset(**self.get_base_kwargs(self))

        if remove_columns:
            new_hf_dataset = new_hf_dataset.remove_columns(column_names=list(removed_field_names))
        else:
            hidden_columns.update(removed_field_names)

        rename_targets_already_in_columns = (
            set(field_mapping.values()) - set(field_mapping)
        ) & hidden_columns
        if len(rename_targets_already_in_columns) > 0:
            raise ValueError(
                f"rename targets are already in column names: {rename_targets_already_in_columns}. Did you "
                f"miss to set remove_columns=True in a previous call of cast_document_type?"
            )

        new_hf_dataset = new_hf_dataset.rename_columns(column_mapping=field_mapping)

        new_dataset = IterableDataset.from_hf_dataset(
            new_hf_dataset, hidden_columns=hidden_columns, document_type=new_document_type
        )

        return new_dataset

    def take(self, n) -> "IterableDataset":
        return self.apply_hf_func(datasets.IterableDataset.take, n=n)



from .builder import GeneratorBasedBuilder
from .dataset import Dataset, IterableDataset
from .dataset_dict import DatasetDict
from .dataset_formatter import DocumentFormatter

__all__ = [
    "GeneratorBasedBuilder",
    "Dataset",
    "IterableDataset",
    "DatasetDict",
    "DocumentFormatter",
]



"""TODO: Add a description here."""


import json
import os

import datasets

_CITATION = """\
@inproceedings{zhang-etal-2017-position,
    title = "Position-aware Attention and Supervised Data Improve Slot Filling",
    author = "Zhang, Yuhao  and
      Zhong, Victor  and
      Chen, Danqi  and
      Angeli, Gabor  and
      Manning, Christopher D.",
    booktitle = "Proceedings of the 2017 Conference on Empirical Methods in Natural Language Processing",
    month = sep,
    year = "2017",
    address = "Copenhagen, Denmark",
    publisher = "Association for Computational Linguistics",
    url = "https://www.aclweb.org/anthology/D17-1004",
    doi = "10.18653/v1/D17-1004",
    pages = "35--45",
}

@inproceedings{alt-etal-2020-tacred,
    title = "{TACRED} Revisited: A Thorough Evaluation of the {TACRED} Relation Extraction Task",
    author = "Alt, Christoph  and
      Gabryszak, Aleksandra  and
      Hennig, Leonhard",
    booktitle = "Proceedings of the 58th Annual Meeting of the Association for Computational Linguistics",
    month = jul,
    year = "2020",
    address = "Online",
    publisher = "Association for Computational Linguistics",
    url = "https://www.aclweb.org/anthology/2020.acl-main.142",
    doi = "10.18653/v1/2020.acl-main.142",
    pages = "1558--1569",
}
"""

# TODO: Add description of the dataset here
# You can copy an official description
_DESCRIPTION = """\
This new dataset is designed to solve this great NLP task and is crafted with a lot of care.
"""

# TODO: Add a link to an official homepage for the dataset here
_HOMEPAGE = ""

# TODO: Add the licence for the dataset here if you can find it
_LICENSE = ""

# TODO: Add link to the official dataset URLs here
# The HuggingFace dataset library don't host the datasets but only point to the original files
# This can be an arbitrary nested dict/list of URLs (see below in `_split_generators` method)
_PATCH_URLs = {
    "dev": "https://raw.githubusercontent.com/DFKI-NLP/tacrev/master/patch/dev_patch.json",
    "test": "https://raw.githubusercontent.com/DFKI-NLP/tacrev/master/patch/test_patch.json",
}

_CLASS_LABELS = [
    "no_relation",
    "org:alternate_names",
    "org:city_of_headquarters",
    "org:country_of_headquarters",
    "org:dissolved",
    "org:founded",
    "org:founded_by",
    "org:member_of",
    "org:members",
    "org:number_of_employees/members",
    "org:parents",
    "org:political/religious_affiliation",
    "org:shareholders",
    "org:stateorprovince_of_headquarters",
    "org:subsidiaries",
    "org:top_members/employees",
    "org:website",
    "per:age",
    "per:alternate_names",
    "per:cause_of_death",
    "per:charges",
    "per:children",
    "per:cities_of_residence",
    "per:city_of_birth",
    "per:city_of_death",
    "per:countries_of_residence",
    "per:country_of_birth",
    "per:country_of_death",
    "per:date_of_birth",
    "per:date_of_death",
    "per:employee_of",
    "per:origin",
    "per:other_family",
    "per:parents",
    "per:religion",
    "per:schools_attended",
    "per:siblings",
    "per:spouse",
    "per:stateorprovince_of_birth",
    "per:stateorprovince_of_death",
    "per:stateorprovinces_of_residence",
    "per:title",
]


def convert_ptb_token(token: str) -> str:
    """Convert PTB tokens to normal tokens"""
    return {
        "-lrb-": "(",
        "-rrb-": ")",
        "-lsb-": "[",
        "-rsb-": "]",
        "-lcb-": "{",
        "-rcb-": "}",
    }.get(token.lower(), token)


# TODO: Name of the dataset usually match the script name with CamelCase instead of snake_case
class TACRED(datasets.GeneratorBasedBuilder):
    """TODO: Short description of my dataset."""

    # This is an example of a dataset with multiple configurations.
    # If you don't want/need to define several sub-sets in your dataset,
    # just remove the BUILDER_CONFIG_CLASS and the BUILDER_CONFIGS attributes.

    # If you need to make complex sub-parts in the datasets with configurable options
    # You can create your own builder configuration class to store attribute, inheriting from datasets.BuilderConfig
    # BUILDER_CONFIG_CLASS = MyBuilderConfig

    # You will be able to load one or the other configurations in the following list with
    # data = datasets.load_dataset('my_dataset', 'first_domain')
    # data = datasets.load_dataset('my_dataset', 'second_domain')
    BUILDER_CONFIGS = [
        datasets.BuilderConfig(
            name="original", version=datasets.Version("1.0.0"), description="The original TACRED."
        ),
        datasets.BuilderConfig(
            name="revised",
            version=datasets.Version("1.0.0"),
            description="The revised TACRED (corrected labels in dev and test split).",
        ),
    ]

    DEFAULT_CONFIG_NAME = "original"  # type: ignore

    @property
    def manual_download_instructions(self):
        return (
            "To use TACRED you have to download it manually. "
            "It is available via the LDC at https://catalog.ldc.upenn.edu/LDC2018T24"
            "Please extract all files in one folder and load the dataset with: "
            "`datasets.load_dataset('tacred', data_dir='path/to/folder/folder_name')`"
        )

    def _info(self):
        features = datasets.Features(
            {
                "tokens": datasets.Sequence(datasets.Value("string")),
                "head_start": datasets.Value("int32"),
                "head_end": datasets.Value("int32"),
                "tail_start": datasets.Value("int32"),
                "tail_end": datasets.Value("int32"),
                "label": datasets.ClassLabel(names=_CLASS_LABELS),
            }
        )

        return datasets.DatasetInfo(
            # This is the description that will appear on the datasets page.
            description=_DESCRIPTION,
            # This defines the different columns of the dataset and their types
            features=features,  # Here we define them above because they are different between the two configurations
            # If there's a common (input, target) tuple from the features,
            # specify them here. They'll be used if as_supervised=True in
            # builder.as_dataset.
            supervised_keys=None,
            # Homepage of the dataset for documentation
            homepage=_HOMEPAGE,
            # License for the dataset if available
            license=_LICENSE,
            # Citation for the dataset
            citation=_CITATION,
        )

    def _split_generators(self, dl_manager):
        """Returns SplitGenerators."""
        # TODO: This method is tasked with downloading/extracting the data and defining the splits depending on the configuration
        # If several configurations are possible (listed in BUILDER_CONFIGS), the configuration selected by the user is in self.config.name

        # dl_manager is a datasets.download.DownloadManager that can be used to download and extract URLs
        # It can accept any type or nested list/dict and will give back the same structure with the url replaced with path to local files.
        # By default the archives will be extracted and a path to a cached folder where they are extracted is returned instead of the archive
        patch_files = {}
        if self.config.name == "revised":
            patch_files = dl_manager.download_and_extract(_PATCH_URLs)

        data_dir = os.path.abspath(os.path.expanduser(dl_manager.manual_dir))

        if not os.path.exists(data_dir):
            raise FileNotFoundError(
                "{} does not exist. Make sure you insert a manual dir via `datasets.load_dataset('tacred', data_dir=...)` that includes the unzipped files from the TACRED_LDC zip. Manual download instructions: {}".format(
                    data_dir, self.manual_download_instructions
                )
            )

        return [
            datasets.SplitGenerator(
                name=datasets.Split.TRAIN,
                gen_kwargs={
                    "filepath": os.path.join(data_dir, "train.json"),
                    "patch_filepath": None,
                },
            ),
            datasets.SplitGenerator(
                name=datasets.Split.TEST,
                gen_kwargs={
                    "filepath": os.path.join(data_dir, "test.json"),
                    "patch_filepath": patch_files.get("test"),
                },
            ),
            datasets.SplitGenerator(
                name=datasets.Split.VALIDATION,
                gen_kwargs={
                    "filepath": os.path.join(data_dir, "dev.json"),
                    "patch_filepath": patch_files.get("dev"),
                },
            ),
        ]

    def _generate_examples(self, filepath, patch_filepath):
        """Yields examples."""
        # TODO: This method will receive as arguments the `gen_kwargs` defined in the previous `_split_generators` method.
        # It is in charge of opening the given file and yielding (key, example) tuples from the dataset
        # The key is not important, it's more here for legacy reason (legacy from tfds)
        patch_examples = {}
        if patch_filepath is not None:
            with open(patch_filepath, encoding="utf-8") as f:
                patch_examples = {example["id"]: example for example in json.load(f)}

        with open(filepath, encoding="utf-8") as f:
            data = json.load(f)
            for example in data:
                id_ = example["id"]

                if id_ in patch_examples:
                    example.update(patch_examples[id_])

                yield id_, {
                    "tokens": [convert_ptb_token(token) for token in example["token"]],
                    "head_start": example["subj_start"],
                    "head_end": example["subj_end"] + 1,  # make end offset exclusive
                    "tail_start": example["obj_start"],
                    "tail_end": example["obj_end"] + 1,  # make end offset exclusive
                    "label": example["relation"],
                }



import dataclasses
from typing import Any, Dict, Optional, Tuple

from typing_extensions import TypeAlias

from pytorch_ie.core import Document


@dataclasses.dataclass
class WithMetadata:
    id: Optional[str] = None
    metadata: Dict[str, Any] = dataclasses.field(default_factory=dict)


@dataclasses.dataclass
class WithTokens:
    tokens: Tuple[str, ...]


@dataclasses.dataclass
class WithText:
    text: str


@dataclasses.dataclass
class TextBasedDocument(WithMetadata, WithText, Document):
    pass


@dataclasses.dataclass
class TokenBasedDocument(WithMetadata, WithTokens, Document):
    pass


# backwards compatibility
TextDocument: TypeAlias = TextBasedDocument



from abc import ABC, abstractmethod
from typing import Optional, Union

from .dataset import Dataset, IterableDataset


class EnterDatasetMixin(ABC):
    """Mixin for processors that enter a dataset context."""

    @abstractmethod
    def enter_dataset(
        self, dataset: Union[Dataset, IterableDataset], name: Optional[str] = None
    ) -> None:
        """Enter dataset context."""


class ExitDatasetMixin(ABC):
    """Mixin for processors that exit a dataset context."""

    @abstractmethod
    def exit_dataset(
        self, dataset: Union[Dataset, IterableDataset], name: Optional[str] = None
    ) -> None:
        """Exit dataset context."""


class EnterDatasetDictMixin(ABC):
    """Mixin for processors that enter a dataset dict context."""

    @abstractmethod
    def enter_dataset_dict(self, dataset_dict) -> None:
        """Enter dataset dict context."""


class ExitDatasetDictMixin(ABC):
    """Mixin for processors that exit a dataset dict context."""

    @abstractmethod
    def exit_dataset_dict(self, dataset_dict) -> None:
        """Exit dataset dict context."""



import dataclasses
import typing
from collections import defaultdict
from collections.abc import Mapping, Sequence
from dataclasses import _asdict_inner  # type: ignore
from typing import (
    Any,
    ClassVar,
    Dict,
    Iterable,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    overload,
)


def _enumerate_dependencies(
    resolved: List[str],
    dependency_graph: Dict[str, List[str]],
    nodes: List[str],
    current_path: Optional[Set[str]] = None,
):
    current_path = current_path or set()
    for node in nodes:
        if node in current_path:
            raise ValueError(f"circular dependency detected at node: {node}")
        if node not in resolved:
            # terminal nodes
            if node not in dependency_graph:
                resolved.append(node)
            # nodes with dependencies
            else:
                # enumerate all dependencies first, then append itself
                _enumerate_dependencies(
                    resolved,
                    dependency_graph,
                    nodes=dependency_graph[node],
                    current_path=current_path | {node},
                )
                resolved.append(node)


def _is_optional_type(t: typing.Type) -> bool:
    type_origin = typing.get_origin(t)
    type_args = typing.get_args(t)
    return type_origin is typing.Union and len(type_args) == 2 and type(None) in type_args


def _is_optional_annotation_type(t: typing.Type) -> bool:
    is_optional = _is_optional_type(t)
    if not is_optional:
        return False
    return _is_annotation_type(typing.get_args(t)[0])


def _is_annotation_type(t: Any) -> bool:
    return type(t) == type and issubclass(t, Annotation)


def _contains_annotation_type(t: Any) -> bool:
    if _is_annotation_type(t):
        return True
    type_args = typing.get_args(t)
    return any(_contains_annotation_type(type_arg) for type_arg in type_args)


def _is_tuple_of_annotation_types(t: Any) -> bool:
    type_args = typing.get_args(t)
    if typing.get_origin(t) == tuple and _is_annotation_type(type_args[0]):
        if not (
            type_args[1] == Ellipsis
            or all(issubclass(type_arg, Annotation) for type_arg in type_args)
        ):
            raise TypeError(
                f"only tuples that do not mix Annotations with other types are supported"
            )
        return True
    else:
        return False


def _get_reference_fields_and_container_types(
    annotation_class: typing.Type["Annotation"],
) -> Dict[str, Any]:
    containers: Dict[str, Any] = {}
    for field in dataclasses.fields(annotation_class):
        if field.name == "_targets":
            continue
        if not _contains_annotation_type(field.type):
            continue
        if _is_optional_annotation_type(field.type):
            containers[field.name] = typing.Optional
            continue
        if _is_annotation_type(field.type):
            containers[field.name] = None
            continue
        if _is_tuple_of_annotation_types(field.type):
            containers[field.name] = tuple
            continue
        annot_name = annotation_class.__name__
        raise TypeError(
            f"The type '{field.type}' of the field '{field.name}' from Annotation subclass '{annot_name}' can not "
            f"be handled automatically. For automatic handling, type constructs that contain any Annotation subclasses "
            f"need to be either (1) pure subclasses of Annotation, (2) tuples of Annotation subclasses, or their "
            f"optional variants (examples: 1) Span, 2) Tuple[Span, ...], 3) Optional[Span]). Is the defined type "
            f"really the one you want to use? If so, consider to overwrite "
            f"{annot_name}.asdict() and {annot_name}.fromdict() by your own."
        )

    return containers


def _get_annotation_fields(fields: List[dataclasses.Field]) -> Set[dataclasses.Field]:
    return {field for field in fields if typing.get_origin(field.type) is AnnotationList}


def _revert_annotation_graph(annotation_graph: Dict[str, List[str]]) -> Dict[str, Set[str]]:
    """Points from field names to the names of dependent annotation fields."""
    reverted_annotation_graph = defaultdict(set)
    all_target_names = set()
    all_dependent_names = set()
    for annotation_name, target_names in annotation_graph.items():
        all_target_names.update(target_names)
        all_dependent_names.add(annotation_name)
        if annotation_name == "_artificial_root":
            continue
        for target_name in target_names:
            reverted_annotation_graph[target_name].add(annotation_name)

    reverted_annotation_graph["_artificial_root"] = all_target_names - all_dependent_names

    return dict(reverted_annotation_graph)


def annotation_field(
    target: Optional[str] = None,
    targets: Optional[List[str]] = None,
    named_targets: Optional[Dict[str, str]] = None,
):
    """
    We allow 3 variants to pass targets:
    1) as single value `target: str`: this works if only one target is required,
    2) as list of target names `targets: List[str]`: this works if multiple targets are required, but the
        respective Annotation class does _not_ define `TARGET_NAMES` (this disallows the usage of named_targets
        in the Annotation class), and
    3) as a mapping `named_targets: Dict[str, str]` from entries in Annotation.TARGET_NAMES to field names of the
        Document: This should be used if the respective Annotation class defines `TARGET_NAMES`, and thus,
        makes use of the `named_targets` property.
    """
    target_names = None
    new_targets = []
    if target is not None:
        if targets is not None or named_targets is not None:
            raise ValueError(f"only one of target, targets or named_targets can be set")
        new_targets = [target]
    if targets is not None:
        if target is not None or named_targets is not None:
            raise ValueError(f"only one of target, targets or named_targets can be set")
        new_targets = targets
    if named_targets is not None:
        if target is not None or targets is not None:
            raise ValueError(f"only one of target, targets or named_targets can be set")
        new_targets = list(named_targets.values())
        target_names = list(named_targets.keys())
    return dataclasses.field(
        metadata=dict(targets=new_targets, target_names=target_names), init=False, repr=False
    )


# for now, we only have annotation lists and texts
TARGET_TYPE = Union["AnnotationList", str]


@dataclasses.dataclass(eq=True, frozen=True)
class Annotation:
    _targets: Optional[Tuple[TARGET_TYPE, ...]] = dataclasses.field(
        default=None, init=False, repr=False, hash=False
    )
    TARGET_NAMES: ClassVar[Optional[Tuple[str, ...]]] = None

    def set_targets(self, value: Optional[Tuple[TARGET_TYPE, ...]]):
        if value is not None and self._targets is not None:
            raise ValueError(
                f"Annotation already has assigned targets. Clear the "
                f"annotation list container or remove the annotation with pop() "
                f"to assign it to a new annotation list with other targets."
            )
        object.__setattr__(self, "_targets", value)

    @property
    def non_comparison_fields_and_values(self) -> Tuple[Tuple[str, Any], ...]:
        return tuple(
            (f.name, getattr(self, f.name)) for f in dataclasses.fields(self) if not f.compare
        )

    @property
    def _id(self) -> int:
        # also calculate the hash over the non-comparison fields for id creation because otherwise
        # these fields would not be considered (per default, hash=false for non-comparison fields)
        # and (de-)serialization would de-duplicate annotations that vary just in these fields
        return hash((self, self.non_comparison_fields_and_values))

    @property
    def target(self) -> Optional[TARGET_TYPE]:
        if self._targets is None:
            return None
        if len(self._targets) == 0:
            raise ValueError(f"annotation has no target")
        if len(self._targets) > 1:
            raise ValueError(
                f"annotation has multiple targets, target is not defined in this case"
            )
        return self._targets[0]

    @property
    def targets(self) -> Optional[Tuple[TARGET_TYPE, ...]]:
        return self._targets

    @property
    def named_targets(self) -> Dict[str, TARGET_TYPE]:
        if self._targets is None:
            raise ValueError(
                f"targets is not set (this annotation may be not yet attached to a document)"
            )
        if self.TARGET_NAMES is None:
            raise TypeError(f"no TARGET_NAMES defined")
        return {name: self._targets[i] for i, name in enumerate(self.TARGET_NAMES)}

    def _asdict(
        self,
        exclude_fields: Optional[List[str]] = None,
        overrides: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        result: List[Tuple[str, Any]] = []
        _exclude_fields = set(exclude_fields) if exclude_fields is not None else set()
        _exclude_fields.add("_targets")
        if overrides is not None:
            _exclude_fields.update(overrides)
            result.extend(overrides.items())
        for f in dataclasses.fields(self):
            if f.name in _exclude_fields:
                continue
            field_value = getattr(self, f.name)
            value = _asdict_inner(field_value, dict)
            result.append((f.name, value))
        dct = dict(result)
        dct["_id"] = self._id
        return dct

    def asdict(self) -> Dict[str, Any]:
        overrides = {}
        reference_fields_with_container_type = _get_reference_fields_and_container_types(
            type(self)
        )
        for field_name, container_type in reference_fields_with_container_type.items():
            if container_type is None:
                overrides[field_name] = getattr(self, field_name)._id
            elif container_type == typing.Optional:
                field_value = getattr(self, field_name)
                overrides[field_name] = None if field_value is None else field_value._id
            elif container_type == tuple:
                # save as list to be json compatible
                overrides[field_name] = [anno._id for anno in getattr(self, field_name)]
            else:
                raise Exception(f"unknown annotation container type: {container_type}")

        dct = self._asdict(overrides=overrides)
        return dct

    @classmethod
    def fromdict(
        cls,
        dct: Dict[str, Any],
        annotation_store: Optional[Dict[int, "Annotation"]] = None,
    ):
        tmp_dct = dict(dct)
        reference_fields_with_container_type = _get_reference_fields_and_container_types(cls)
        for field_name, container_type in reference_fields_with_container_type.items():
            if container_type is None:
                tmp_dct[field_name] = resolve_annotation(
                    tmp_dct[field_name], store=annotation_store
                )
            elif container_type == typing.Optional:
                if tmp_dct[field_name] is None:
                    tmp_dct[field_name] = None
                else:
                    tmp_dct[field_name] = resolve_annotation(
                        tmp_dct[field_name], store=annotation_store
                    )
            elif container_type == tuple:
                tmp_dct[field_name] = tuple(
                    resolve_annotation(anno_dct, store=annotation_store)
                    for anno_dct in tmp_dct[field_name]
                )
            else:
                raise Exception(f"unknown annotation container type: {container_type}")

        tmp_dct.pop("_id", None)
        return cls(**tmp_dct)

    @property
    def is_attached(self) -> bool:
        return self._targets is not None

    def copy(self, **overrides) -> "Annotation":
        """
        Create a detached copy of the annotation with the same values as the original.

        :param overrides: keyword arguments to override the values of the original annotation
        :return: a detached copy of the annotation
        """
        kwargs = {}
        for f in dataclasses.fields(self):
            if f.name == "_targets":
                continue
            kwargs[f.name] = getattr(self, f.name)
        kwargs.update(overrides)
        return type(self)(**kwargs)

    def copy_with_store(self, override_annotation_store: Dict[int, "Annotation"]) -> "Annotation":
        """
        Create a detached copy of the annotation, but replace references to other annotations
        with entries from an override annotation store.
        Note: For now, fields are only allowed to directly reference an annotation or a tuple of annotations,
         but not any nested data structure that contains annotations.

        Args:
            :param override_annotation_store: the annotation store to use, a mapping from original *annotation ids*
                to the new *annotations*
            :return: a detached copy of the annotation
        """
        overrides: Dict[str, Any] = {}
        for f in dataclasses.fields(self):
            if f.name == "_targets":
                continue
            field_value = getattr(self, f.name)
            if isinstance(field_value, Annotation):
                overrides[f.name] = override_annotation_store.get(field_value._id, field_value)
            elif isinstance(field_value, tuple):
                overrides[f.name] = tuple(
                    override_annotation_store.get(maybe_anno._id, maybe_anno)
                    if isinstance(maybe_anno, Annotation)
                    else maybe_anno
                    for maybe_anno in field_value
                )
            elif isinstance(field_value, (int, float, str, bool, type(None))):
                continue
            else:
                raise Exception(f"unknown annotation field type: {type(field_value)}")

        return self.copy(**overrides)


T = TypeVar("T", covariant=False, bound="Annotation")


class BaseAnnotationList(Sequence[T]):
    def __init__(self, document: "Document", targets: List[str]):
        self._document = document
        self._targets = targets
        self._annotations: List[T] = []

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseAnnotationList):
            return NotImplemented

        return self._targets == other._targets and self._annotations == other._annotations

    @overload
    def __getitem__(self, index: int) -> T:
        ...

    @overload
    def __getitem__(self, s: slice) -> List[T]:
        ...

    def __getitem__(self, index: Union[int, slice]) -> Union[T, List[T]]:
        return self._annotations[index]

    def __len__(self) -> int:
        return len(self._annotations)

    def append(self, annotation: T) -> None:
        targets = tuple(getattr(self._document, target_name) for target_name in self._targets)
        annotation.set_targets(targets)
        self._annotations.append(annotation)

    def extend(self, annotations: Iterable[T]) -> None:
        for annotation in annotations:
            self.append(annotation)

    def __repr__(self) -> str:
        return f"BaseAnnotationList({str(self._annotations)})"

    def clear(self) -> List[T]:
        """
        Detach all annotations from the layer and return them.
        """
        result = list(self._annotations)
        for annotation in self._annotations:
            annotation.set_targets(None)
        self._annotations = []
        return result

    def pop(self, index: int = -1) -> T:
        ann = self._annotations.pop(index)
        ann.set_targets(None)
        return ann

    @property
    def targets(self) -> dict[str, Any]:
        return {
            target_field_name: getattr(self._document, target_field_name)
            for target_field_name in self._targets
        }

    @property
    def target(self) -> Any:
        tgts = self.targets
        if len(tgts) != 1:
            raise ValueError(
                f"The annotation layer has more or less than one target: {self._targets}"
            )
        return list(tgts.values())[0]

    @property
    def target_layers(self) -> dict[str, "AnnotationList"]:
        return {
            target_name: target
            for target_name, target in self.targets.items()
            if isinstance(target, AnnotationList)
        }

    @property
    def target_layer(self) -> "AnnotationList":
        tgt_layers = self.target_layers
        if len(tgt_layers) != 1:
            raise ValueError(
                f"The annotation layer has more or less than one target layer: {list(tgt_layers.keys())}"
            )
        return list(tgt_layers.values())[0]


class AnnotationList(BaseAnnotationList[T]):
    def __init__(self, document: "Document", targets: List["str"]):
        super().__init__(document=document, targets=targets)
        self._predictions: BaseAnnotationList[T] = BaseAnnotationList(document, targets=targets)

    @property
    def predictions(self) -> BaseAnnotationList[T]:
        return self._predictions

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, AnnotationList):
            return NotImplemented

        return super().__eq__(other) and self.predictions == other.predictions

    def __repr__(self) -> str:
        return f"AnnotationList({str(self._annotations)})"


D = TypeVar("D", bound="Document")


@dataclasses.dataclass
class Document(Mapping[str, Any]):
    # points from annotation field names to lists of target field names
    _annotation_graph: Dict[str, List[str]] = dataclasses.field(
        default_factory=dict, init=False, repr=False
    )
    _annotation_fields: Set[str] = dataclasses.field(default_factory=set, init=False, repr=False)

    @classmethod
    def fields(cls):
        return [
            f
            for f in dataclasses.fields(cls)
            if f.name not in {"_annotation_graph", "_annotation_fields"}
        ]

    @classmethod
    def annotation_fields(cls):
        return _get_annotation_fields(list(dataclasses.fields(cls)))

    def __getitem__(self, key: str) -> AnnotationList:
        if key not in self._annotation_fields:
            raise KeyError(f"Document has no attribute '{key}'.")
        return getattr(self, key)

    def __iter__(self):
        return iter(self._annotation_fields)

    def __len__(self):
        return len(self._annotation_fields)

    def __post_init__(self):
        targeted = set()
        field_names = {field.name for field in dataclasses.fields(self)}
        for field in dataclasses.fields(self):
            if field.name == "_annotation_graph":
                continue

            field_origin = typing.get_origin(field.type)

            if field_origin is AnnotationList:
                self._annotation_fields.add(field.name)

                targets = field.metadata.get("targets")
                for target in targets:
                    targeted.add(target)
                    if field.name not in self._annotation_graph:
                        self._annotation_graph[field.name] = []
                    self._annotation_graph[field.name].append(target)
                    if target not in field_names:
                        raise TypeError(
                            f'annotation target "{target}" is not in field names of the document: {field_names}'
                        )

                # check annotation target names and use them together with target names from the AnnotationList
                # to reorder targets, if available
                target_names = field.metadata.get("target_names")
                annotation_type = typing.get_args(field.type)[0]
                annotation_target_names = annotation_type.TARGET_NAMES
                if annotation_target_names is not None:
                    if target_names is not None:
                        if set(target_names) != set(annotation_target_names):
                            raise TypeError(
                                f"keys of targets {sorted(target_names)} do not match "
                                f"{annotation_type.__name__}.TARGET_NAMES {sorted(annotation_target_names)}"
                            )
                        # reorder targets according to annotation_target_names
                        target_name_mapping = dict(zip(target_names, targets))
                        target_position_mapping = {
                            i: target_name_mapping[name]
                            for i, name in enumerate(annotation_target_names)
                        }
                        targets = [target_position_mapping[i] for i in range(len(targets))]
                    else:
                        if len(annotation_target_names) != len(targets):
                            raise TypeError(
                                f"number of targets {sorted(targets)} does not match number of entries in "
                                f"{annotation_type.__name__}.TARGET_NAMES: {sorted(annotation_target_names)}"
                            )
                        # disallow multiple targets when target names are specified in the definition of the Annotation
                        if len(annotation_target_names) > 1:
                            raise TypeError(
                                f"A target name mapping is required for AnnotationLists containing Annotations with "
                                f'TARGET_NAMES, but AnnotationList "{field.name}" has no target_names. You should '
                                f"pass the named_targets dict containing the following keys (see Annotation "
                                f'"{annotation_type.__name__}") to annotation_field: {annotation_target_names}'
                            )

                field_value = field.type(document=self, targets=targets)
                setattr(self, field.name, field_value)

        if "_artificial_root" in self._annotation_graph:
            raise ValueError(
                'Failed to add the "_artificial_root" node to the annotation graph because it already exists. Note '
                "that AnnotationList entries with that name are not allowed."
            )
        self._annotation_graph["_artificial_root"] = list(self._annotation_fields - targeted)

    def asdict(self):
        dct = {}
        for field in self.fields():
            value = getattr(self, field.name)

            if isinstance(value, AnnotationList):
                dct[field.name] = {
                    "annotations": [v.asdict() for v in value],
                    "predictions": [v.asdict() for v in value.predictions],
                }
            elif isinstance(value, dict):
                dct[field.name] = value or None
            else:
                dct[field.name] = value

        return dct

    @classmethod
    def fromdict(cls, dct):
        fields = dataclasses.fields(cls)
        annotation_fields = _get_annotation_fields(fields)

        cls_kwargs = {}
        for field in fields:
            if field not in annotation_fields:
                value = dct.get(field.name)

                if value is not None:
                    cls_kwargs[field.name] = value

        doc = cls(**cls_kwargs)

        name_to_field = {f.name: f for f in annotation_fields}

        dependency_ordered_fields: List[str] = []
        _enumerate_dependencies(
            dependency_ordered_fields,
            dependency_graph=doc._annotation_graph,
            nodes=doc._annotation_graph["_artificial_root"],
        )

        annotations = {}
        predictions = {}
        annotations_per_field = defaultdict(list)
        predictions_per_field = defaultdict(list)
        for field_name in dependency_ordered_fields:
            # terminal nodes do not have to be an annotation field (e.g. the text field)
            if field_name not in name_to_field:
                continue

            field = name_to_field[field_name]

            value = dct.get(field.name)

            if value is None or not value:
                continue

            # TODO: handle single annotations, e.g. a document-level label
            if typing.get_origin(field.type) is AnnotationList:
                annotation_class = typing.get_args(field.type)[0]
                # build annotations
                for annotation_data in value["annotations"]:
                    annotation_dict = dict(annotation_data)
                    annotation_id = annotation_dict.pop("_id")
                    # annotations can only reference annotations
                    annotation = annotation_class.fromdict(annotation_dict, annotations)
                    annotations[annotation_id] = annotation
                    annotations_per_field[field.name].append(annotation)
                # build predictions
                for annotation_data in value["predictions"]:
                    annotation_dict = dict(annotation_data)
                    annotation_id = annotation_dict.pop("_id")
                    # predictions can reference annotations and predictions
                    annotation = annotation_class.fromdict(
                        annotation_dict, {**annotations, **predictions}
                    )
                    predictions[annotation_id] = annotation
                    predictions_per_field[field.name].append(annotation)
            else:
                raise Exception("Error")

        for field_name, field_annotations in annotations_per_field.items():
            getattr(doc, field_name).extend(field_annotations)

        for field_name, field_annotations in predictions_per_field.items():
            getattr(doc, field_name).predictions.extend(field_annotations)

        return doc

    def as_type(
        self,
        new_type: typing.Type[D],
        field_mapping: Optional[Dict[str, str]] = None,
        keep_remaining: bool = True,
    ) -> D:
        field_mapping = field_mapping or {}
        new_doc = new_type.fromdict(
            {
                field_mapping.get(k, k): v
                for k, v in self.asdict().items()
                if keep_remaining or k in field_mapping
            }
        )
        return new_doc

    def copy(self) -> "Document":
        return type(self).fromdict(self.asdict())

    def add_all_annotations_from_other(
        self,
        other: "Document",
        override_annotation_mapping: Optional[Dict[str, Dict[int, Annotation]]] = None,
        process_predictions: bool = True,
    ) -> None:
        """Adds all annotations from another document to this document.

        Args:
            other: The document to add annotations from.
            process_predictions: Whether to process predictions as well (default: True). If set to False,
                the predictions in the other document will be ignored.
            override_annotation_mapping: A mapping from annotation field names to a mapping from
                annotation IDs to annotations. The effects are two-fold. First, adding any annotation
                field name as key has the effect that the field is expected to be already handled and
                no annotations will be added from the other document. Second, if a certain mapping
                old_annotation._id -> new_annotation is present in the mapping for a certain field,
                the new_annotation will be used anywhere where the old_annotation would have been
                referenced. This propagates along the annotation graph and can be useful if some
                annotations are modified, but all dependent annotations should be kept intact e.g.
                when converting a text-based document to token-based (or the other way around):

                ```
                @dataclasses.dataclass(frozen=True)
                class Attribute(Annotation):
                    ref: Annotation
                    value: str

                @dataclasses.dataclass
                class TextBasedDocumentWithEntitiesRelationsAndRelationAttributes(TextBasedDocument):
                    entities: AnnotationList[LabeledSpan] = annotation_field(target="text")
                    relations: AnnotationList[BinaryRelation] = annotation_field(target="entities")
                    relation_attributes: AnnotationList[Attribute] = annotation_field(target="relations")

                @dataclasses.dataclass
                class TokenBasedDocumentWithEntitiesRelationsAndRelationAttributes(TokenBasedDocument):
                    entities: AnnotationList[LabeledSpan] = annotation_field(target="tokens")
                    relations: AnnotationList[BinaryRelation] = annotation_field(target="entities")
                    relation_attributes: AnnotationList[Attribute] = annotation_field(target="relations")


                doc_text = TextBasedDocumentWithEntitiesRelationsAndRelationAttributes(text="Hello World!")
                e1_text = LabeledSpan(0, 5, "word1")
                e2_text = LabeledSpan(6, 11, "word2")
                doc_text.entities.extend([e1, e2])
                r1_text = BinaryRelation(e1_text, e2_text, "relation1")
                doc_text.relations.append(r1_text)
                doc_text.relation_attributes.append(Attribute(r1_text, "attribute1"))

                doc_tokens = TokenBasedDocumentWithEntitiesRelationsAndRelationAttributes(
                    tokens=("Hello", "World", "!")
                )
                e1_tokens = LabeledSpan(0, 1, "word1")
                e2_tokens = LabeledSpan(1, 2, "word2")
                doc_tokens.entities.extend([e1_tokens, e2_tokens])
                doc_tokens.add_all_annotations_from_other(
                    other=doc_text,
                    override_annotation_mapping={"entities": {e1_text._id: e1_tokens, e2_text._id: e2_tokens}},
                )
                # Note that the relation and attribute are still present, but now refer to the new entities
                # and new relation, respectively.
                ```
        """

        annotation_store: Dict[str, Dict[int, Annotation]] = defaultdict(dict)
        named_annotation_fields = {field.name: field for field in self.annotation_fields()}
        if override_annotation_mapping is not None:
            for field_name, mapping in override_annotation_mapping.items():
                if field_name not in named_annotation_fields:
                    raise ValueError(
                        f'Field "{field_name}" is not an annotation field of {type(self).__name__}, but keys in '
                        f"override_annotation_mapping must be annotation field names."
                    )
                annotation_store[field_name].update(mapping)
        else:
            override_annotation_mapping = dict()

        fields_todo = ["_artificial_root"]
        fields_done = set()
        reverted_annotation_graph = _revert_annotation_graph(self._annotation_graph)
        while fields_todo:
            field_name = fields_todo.pop(0)
            if field_name not in fields_done:
                if (
                    field_name in named_annotation_fields
                    and field_name not in override_annotation_mapping
                ):
                    current_targets_store = dict()
                    for target in self._annotation_graph.get(field_name, []):
                        current_targets_store.update(annotation_store[target])

                    other_annotation_field = other[field_name]
                    for ann in other_annotation_field:
                        new_ann = ann.copy_with_store(current_targets_store)
                        if ann._id != new_ann._id:
                            annotation_store[field_name][ann._id] = new_ann
                        self[field_name].append(new_ann)
                    if process_predictions:
                        for ann in other_annotation_field.predictions:
                            new_ann = ann.copy_with_store(current_targets_store)
                            if ann._id != new_ann._id:
                                annotation_store[field_name][ann._id] = new_ann
                            self[field_name].predictions.append(new_ann)
                fields_todo.extend(reverted_annotation_graph.get(field_name, []))
                fields_done.add(field_name)


def resolve_annotation(
    id_or_annotation: Union[int, Annotation],
    store: Optional[Dict[int, Annotation]],
) -> Annotation:
    if isinstance(id_or_annotation, Annotation):
        return id_or_annotation
    else:
        if store is None:
            raise ValueError("Unable to resolve the annotation id without annotation_store.")
        return store[id_or_annotation]



"""TODO: Add a description here."""


import os
import re

import tensorflow as tf
from spacy.lang.en import English

import datasets

_CITATION_WEBRED = """\
@misc{ormandi2021webred,
    title={WebRED: Effective Pretraining And Finetuning For Relation Extraction On The Web},
    author={Robert Ormandi and Mohammad Saleh and Erin Winter and Vinay Rao},
    year={2021},
    eprint={2102.09681},
    archivePrefix={arXiv},
    primaryClass={cs.CL},
    url={https://arxiv.org/abs/2102.09681},
}
"""

# You can copy an official description
_DESCRIPTION = """\
A dataset for extracting relationships from a variety of text found on the World Wide Web. Text
on the web has diverse surface forms including writing styles, complexity and grammar. This dataset collects
sentences from a variety of webpages and documents that represent a variety of those categories. In each sentence,
there will be a subject and object entities tagged with subject SUBJ{...} and object OBJ{...}, respectively. The two
entities are either related by a relation from a set of pre-defined ones or has no relation.

More information about the dataset can be found in our paper: https://arxiv.org/abs/2102.09681
"""

_HOMEPAGE = "https://github.com/google-research-datasets/WebRED"

_LICENSE = """\
This data is licensed by Google LLC under a Creative Commons Attribution 4.0 International License (
http://creativecommons.org/licenses/by/4.0/) Users will be allowed to modify and repost it, and we encourage them to
analyze and publish research based on the data.
"""

# The HuggingFace dataset library don't host the datasets but only point to the original files
# This can be an arbitrary nested dict/list of URLs (see below in `_split_generators` method)
_DATA_URL = "https://github.com/google-research-datasets/WebRED"

_CLASS_LABELS = [
    "based on",
    "part of the series",
    "drug used for treatment",
    "architectural style",
    "writable file format",
    "work location",
    "position held",
    "followed by",
    "flash point",
    "indigenous to",
    "Mohs' hardness",
    "political alignment",
    "located in protected area",
    "translator",
    "director",
    "highest judicial authority",
    "producer",
    "compressive modulus of elasticity",
    "series spin-off",
    "quantity",
    "lyrics by",
    "cell component",
    "medical condition treated",
    "place of death",
    "number of seats",
    "record label",
    "league level above",
    "military branch",
    "origin of the watercourse",
    "diameter",
    "conversion to SI unit",
    "works in collection",
    "presenter",
    "chairperson",
    "temperature",
    "currency",
    "frequency",
    "standards body",
    "manufacturer",
    "location of final assembly",
    "coat of arms",
    "astronaut mission",
    "length",
    "publication date",
    "place of publication",
    "country of citizenship",
    "minimal lethal dose",
    "game mechanics",
    "afflicts",
    "used by",
    "oxidation state",
    "mother",
    "affiliation",
    "head of state",
    "creator",
    "defendant",
    "head coach of sports team",
    "country",
    "developer",
    "approved by",
    "cover artist",
    "lake inflows",
    "separated from",
    "operating area",
    "water as percent of area",
    "head coach",
    "update method",
    "floruit",
    "party chief representative",
    "commander of",
    "gestation period",
    "religious order",
    "school district",
    "depicted by",
    "publisher",
    "excavation director",
    "airline alliance",
    "librettist",
    "executive producer",
    "donated by",
    "mushroom ecological type",
    "iconographic symbol",
    "speed limit",
    "number of representatives in an organization/legislature",
    "subsidiary",
    "educated at",
    "number of participants",
    "founded by",
    "country of origin",
    "family",
    "package management system",
    "subject has role",
    "sibling",
    "interchange station",
    "facet of",
    "decays to",
    "repeals",
    "legislative body",
    "occupant",
    "atomic number",
    "CPU",
    "GUI toolkit or framework",
    "has parts of the class",
    "director of photography",
    "shares border with",
    "parent organization",
    "population",
    "upper flammable limit",
    "performer",
    "isospin z-component",
    "number of injured",
    "number of seasons",
    "choreographer",
    "replaces",
    "doctoral advisor",
    "official residence",
    "top-level Internet domain",
    "VAT-rate",
    "point in time",
    "distance from Earth",
    "public holiday",
    "languages spoken, written or signed",
    "located on astronomical location",
    "solved by",
    "designed by",
    "twinned administrative body",
    "encoded by",
    "located in time zone",
    "canonization status",
    "date of official opening",
    "student",
    "brand",
    "refractive index",
    "inflation rate",
    "home venue",
    "neutron number",
    "chief operating officer",
    "lowest point",
    "signatory",
    "consecrator",
    "model item",
    "time of earliest written record",
    "area",
    "terminus location",
    "significant event",
    "inspired by",
    "backup or reserve team or crew",
    "maximum number of players",
    "talk show guest",
    "number of deaths",
    "exclave of",
    "maximal incubation period in humans",
    "league",
    "film crew member",
    "electric charge",
    "symptoms",
    "replaced by",
    "nominated for",
    "religion",
    "wavelength",
    "total produced",
    "time of discovery or invention",
    "invasive to",
    "use",
    "negative therapeutic predictor",
    "item operated",
    "participating team",
    "political ideology",
    "compulsory education (maximum age)",
    "applies to jurisdiction",
    "history of topic",
    "author",
    "mass",
    "heart rate",
    "killed by",
    "characters",
    "diocese",
    "Erdős number",
    "time period",
    "has part",
    "age of candidacy",
    "semi-major axis",
    "dual to",
    "official language",
    "production company",
    "replaced synonym (for nom. nov.)",
    "main regulatory text",
    "participant of",
    "head of government",
    "age of majority",
    "heritage designation",
    "drafted by",
    "family relationship degree",
    "discontinued date",
    "operator",
    "term length of office",
    "spin quantum number",
    "vehicles per capita (1000)",
    "enclave within",
    "embodied energy",
    "represents",
    "partner",
    "stepparent",
    "taxon synonym",
    "time of spacecraft launch",
    "conversion to standard unit",
    "nominal GDP",
    "lower flammable limit",
    "readable file format",
    "minimal incubation period in humans",
    "connecting line",
    "located in the administrative territorial entity",
    "place of burial",
    "contains administrative territorial entity",
    "statistical leader",
    "sports discipline competed in",
    "tensile modulus of elasticity",
    "research site",
    "connects with",
    "has cause",
    "date of birth",
    "location",
    "age of consent",
    "mains voltage",
    "industry",
    "basionym",
    "marriageable age",
    "visitors per year",
    "Poisson's ratio",
    "suicide rate",
    "carries scientific instrument",
    "connecting service",
    "place of detention",
    "crew member",
    "place served by transport hub",
    "organisation directed from the office or person",
    "memory capacity",
    "primary destinations",
    "relative permeability",
    "parent club",
    "organizer",
    "space launch vehicle",
    "encodes",
    "architect",
    "notable work",
    "commissioned by",
    "depicts",
    "individual tax rate",
    "website account on",
    "central bank",
    "software engine",
    "numeric value",
    "official religion",
    "wingspan",
    "occupation",
    "member count",
    "ceiling exposure limit",
    "date of first performance",
    "discoverer or inventor",
    "described by source",
    "executive body",
    "parent taxon",
    "pole position",
    "sports league level",
    "pKa",
    "genetic association",
    "mountain range",
    "part of",
    "legal form",
    "regulates (molecular biology)",
    "end time",
    "month of the year",
    "employer",
    "from fictional universe",
    "spouse",
    "copyright holder",
    "lake outflow",
    "solubility",
    "located in or next to body of water",
    "IDLH",
    "office held by head of the organisation",
    "office held by head of government",
    "territory claimed by",
    "tracklist",
    "takes place in fictional universe",
    "mount",
    "season of club or team",
    "this taxon is source of",
    "theme music",
    "Alexa rank",
    "film editor",
    "derivative work",
    "territory overlaps",
    "perimeter",
    "price",
    "secretary general",
    "frequency of event",
    "mascot",
    "maintained by",
    "duration",
    "screenwriter",
    "life expectancy",
    "minimum number of players",
    "winner",
    "native language",
    "start time",
    "highest point",
    "legislated by",
    "parity",
    "melting point",
    "location of formation",
    "ultimate tensile strength",
    "defined daily dose",
    "chief executive officer",
    "number of parts of this work of art",
    "endemic to",
    "subclass of",
    "dissolved, abolished or demolished",
    "service entry",
    "follows",
    "number of constituencies",
    "structural engineer",
    "writing system",
    "capital of",
    "taxonomic type",
    "next higher rank",
    "commemorates",
    "continent",
    "relative",
    "residence time of water",
    "number of speakers",
    "conferred by",
    "Gram staining",
    "work period (start)",
    "sport",
    "has effect",
    "tributary",
    "place of birth",
    "member of sports team",
    "relative permittivity",
    "instrument",
    "interested in",
    "academic degree",
    "location of discovery",
    "electronegativity",
    "located on terrain feature",
    "conflict",
    "height",
    "short-term exposure limit",
    "start point",
    "original language of film or TV show",
    "publication interval",
    "amended by",
    "material used",
    "located in present-day administrative territorial entity",
    "drainage basin",
    "lakes on river",
    "league level below",
    "licensed to broadcast to",
    "residence",
    "after a work by",
    "present in work",
    "basin country",
    "product certification",
    "mouth of the watercourse",
    "for work",
    "has quality",
    "uses",
    "time-weighted average exposure limit",
    "license",
    "significant person",
    "archives at",
    "natural product of taxon",
    "anthem",
    "adjacent station",
    "real gross domestic product growth rate",
    "carries",
    "member of political party",
    "professional or sports partner",
    "ethnic group",
    "member of",
    "platform",
    "destination point",
    "sports season of league or competition",
    "country for sport",
    "account charge / subscription fee",
    "patron saint",
    "compulsory education (minimum age)",
    "route of administration",
    "antiparticle",
    "sponsor",
    "floors above ground",
    "timezone offset",
    "programming language",
    "stock exchange",
    "opposite of",
    "mouthpiece",
    "unemployment rate",
    "watershed area",
    "editor",
    "collection",
    "award received",
    "designated as terrorist by",
    "illustrator",
    "student of",
    "dedicated to",
    "youth wing",
    "total fertility rate",
    "elevation above sea level",
    "repealed by",
    "practiced by",
    "named after",
    "movement",
    "flattening",
    "position played on team / speciality",
    "median lethal dose",
    "employees",
    "physically interacts with",
    "highway system",
    "parent peak",
    "participant",
    "number of cases",
    "editor-in-chief",
    "instance of",
    "sidekick of",
    "width",
    "cites",
    "child",
    "has edition",
    "doctoral student",
    "original network",
    "board member",
    "service retirement",
    "anatomical location",
    "biological variant of",
    "Euler characteristic",
    "diplomatic relation",
    "number of children",
    "narrative location",
    "incidence",
    "allegiance",
    "airline hub",
    "vapor pressure",
    "constellation",
    "voice actor",
    "number of platform tracks",
    "work period (end)",
    "military rank",
    "vertical depth",
    "vessel class",
    "parent astronomical body",
    "director/manager",
    "owner of",
    "distribution",
    "court",
    "angular resolution",
    "located on street",
    "owned by",
    "retirement age",
    "said to be the same as",
    "language used",
    "applies to part",
    "business division",
    "contains settlement",
    "main subject",
    "operating system",
    "authority",
    "number of representations",
    "ancestral home",
    "radius",
    "binding energy",
    "general manager",
    "measured by",
    "next lower rank",
    "cast member",
    "thermal conductivity",
    "health specialty",
    "father",
    "worshipped by",
    "headquarters location",
    "child astronomical body",
    "distributor",
    "noble title",
    "studied by",
    "officeholder",
    "genre",
    "vaccine for",
    "inception",
    "produced by",
    "narrator",
    "different from",
    "volcano observatory",
    "art director",
    "objective of project or action",
    "composer",
    "hardness",
    "edition or translation of",
    "isospin quantum number",
    "foundational text",
    "broadcast by",
    "office held by head of state",
    "boiling point",
    "minimum wavelength of sensitivity",
    "speaker",
    "studies",
    "capital",
    "terminus",
    "pressure",
    "number of episodes",
    "decomposition point",
    "filming location",
    "product or material produced",
    "gene inversion association with",
    "found in taxon",
    "field of work",
    "language of work or name",
    "ranking",
    "crosses",
    "culture",
    "location of first performance",
    "dialect of",
    "date of death",
    "influenced by",
]


class WebRedConfig(datasets.BuilderConfig):
    """BuilderConfig for WebRed."""

    def __init__(
        self,
        data_url,
        citation,
        url,
        class_labels,
        description,
        **kwargs,
    ):
        """BuilderConfig for WebRed.
        Args:
          data_url: `string`, url to download the zip file from
          citation: `string`, citation for the data set
          url: `string`, url for information about the data set
          class_labels: `list[string]`, the list of classes if the label is
            categorical. If not provided, then the label will be of type
            `datasets.Value('float32')`.
          **kwargs: keyword arguments forwarded to super.
        """
        super().__init__(version=datasets.Version("1.0.0", ""), **kwargs)
        self.class_labels = class_labels
        self.data_url = data_url
        self.citation = citation
        self.url = url
        self.description = description


# TODO: Name of the dataset usually match the script name with CamelCase instead of snake_case
class WebRed(datasets.GeneratorBasedBuilder):
    """TODO: Short description of my dataset."""

    BUILDER_CONFIGS = [
        WebRedConfig(
            name="webred_5",
            data_url="https://github.com/google-research-datasets/WebRED/raw/main/webred_5.tfrecord",
            citation=_CITATION_WEBRED,
            url=_HOMEPAGE,
            class_labels=_CLASS_LABELS,
            description=_DESCRIPTION
            + "\nEach example in WebRED 5 was annotated by exactly 5 independent human annotators.",
        ),
        WebRedConfig(
            name="webred_21",
            data_url="https://github.com/google-research-datasets/WebRED/raw/main/webred_21.tfrecord",
            citation=_CITATION_WEBRED,
            url=_HOMEPAGE,
            class_labels=_CLASS_LABELS,
            description=_DESCRIPTION
            + "\nIn WebRED 2+1, each example was annotated by 2 independent annotators. If they "
            "disagreed, an additional annotator (+1) was assigned to the example who also "
            "provided a disambiguating annotation.",
        ),
    ]

    def _info(self):
        return datasets.DatasetInfo(
            # This is the description that will appear on the datasets page.
            description=self.config.description,
            # This defines the different columns of the dataset and their types
            features=datasets.Features(
                {
                    "tokens": datasets.Sequence(datasets.Value("string")),
                    "head_start": datasets.Value("int32"),
                    "head_end": datasets.Value("int32"),
                    "tail_start": datasets.Value("int32"),
                    "tail_end": datasets.Value("int32"),
                    "label": datasets.ClassLabel(names=_CLASS_LABELS),
                }
            ),
            # If there's a common (input, target) tuple from the features,
            # specify them here. They'll be used if as_supervised=True in
            # builder.as_dataset.
            supervised_keys=None,
            # Homepage of the dataset for documentation
            homepage=_HOMEPAGE,
            # License for the dataset if available
            license=_LICENSE,
            # Citation for the dataset
            citation=_CITATION_WEBRED,
        )

    def _split_generators(self, dl_manager):
        """Returns SplitGenerators."""
        # This method is tasked with downloading/extracting the data and defining the splits depending on the configuration
        # If several configurations are possible (listed in BUILDER_CONFIGS), the configuration selected by the user is in self.config.name

        # dl_manager is a datasets.download.DownloadManager that can be used to download and extract URLs
        # It can accept any type or nested list/dict and will give back the same structure with the url replaced with path to local files.
        # By default the archives will be extracted and a path to a cached folder where they are extracted is returned instead of the archive
        file_path = dl_manager.download_and_extract(self.config.data_url)

        return [
            datasets.SplitGenerator(
                name=datasets.Split.TRAIN,
                gen_kwargs={"filepath": file_path},
            ),
        ]

    def _generate_examples(self, filepath):
        """Yields examples."""
        dataset = tf.data.TFRecordDataset(filepath)
        idx = 0
        nlp = English()

        def get_feature_value(feature, key):
            return feature[key].bytes_list.value[0].decode("utf-8")

        for raw_sentence in dataset:
            example = tf.train.Example()
            example.ParseFromString(raw_sentence.numpy())

            rel_id = get_feature_value(example.features.feature, "relation_id")
            sentence = get_feature_value(example.features.feature, "sentence")
            label = get_feature_value(example.features.feature, "relation_name")

            # 1. Find OBJ{} and SUBJ{} marker indices
            subj = re.search("SUBJ{.+?}", sentence)
            obj = re.search("OBJ{.+?}", sentence)
            if not subj or not obj:
                print(f"Did not find OBJ or SUBJ marker in sentence: {sentence}")
                continue
            else:
                subj_start, subj_end = subj.span()
                obj_start, obj_end = obj.span()
            # 2. OPTIONAL: Replace with source and target strings (they contain special characters while the sentence
            # contains standard writing?)
            # source = get_feature_value(sentence.features.feature, "source_name")
            # target = get_feature_value(sentence.features.feature, "target_name")

            # 3. Remove markers and adjust indices: divide sentence at marker indices, remove marker, merge
            # what if subj or obj is at the start or end of the sentence?
            cleaned_sentence = ""
            if subj_start < obj_start:
                cleaned_sentence += sentence[:subj_start]
                cleaned_sentence += sentence[subj_start + 5 : subj_end - 1]
                cleaned_sentence += sentence[subj_end:obj_start]
                cleaned_sentence += sentence[obj_start + 4 : obj_end - 1]
                cleaned_sentence += sentence[obj_end:]
                subj_end -= 6
                obj_start -= 6
                obj_end -= 11
            else:
                cleaned_sentence += sentence[:obj_start]
                cleaned_sentence += sentence[obj_start + 4 : obj_end - 1]
                cleaned_sentence += sentence[obj_end:subj_start]
                cleaned_sentence += sentence[subj_start + 5 : subj_end - 1]
                cleaned_sentence += sentence[subj_end:]
                obj_end -= 5
                subj_start -= 5
                subj_end -= 11
            # 4. Tokenize and calculate token indices from char offsets
            doc = nlp(cleaned_sentence)
            tokens = [token.text for token in doc]
            subj_span = doc.char_span(subj_start, subj_end, alignment_mode="expand")
            head_start = subj_span.start
            head_end = subj_span.end
            obj_span = doc.char_span(obj_start, obj_end, alignment_mode="expand")
            tail_start = obj_span.start
            tail_end = obj_span.end

            id_ = str(idx) + "_" + rel_id
            idx += 1

            yield id_, {
                "tokens": tokens,
                "head_start": head_start,
                "head_end": head_end,
                "tail_start": tail_start,
                "tail_end": tail_end,
                "label": label,
            }



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pytorch_ie',
 'pytorch_ie.core',
 'pytorch_ie.data',
 'pytorch_ie.data.datasets',
 'pytorch_ie.data.datasets.hf_datasets',
 'pytorch_ie.metrics',
 'pytorch_ie.models',
 'pytorch_ie.models.genre',
 'pytorch_ie.models.modules',
 'pytorch_ie.taskmodules',
 'pytorch_ie.utils']

package_data = \
{'': ['*']}

install_requires = \
['absl-py>=1.0.0,<2.0.0',
 'datasets>=2.9,<2.10',
 'huggingface-hub>=0.5.1,<0.6.0',
 'pytorch-lightning>=1.9,<2.0',
 'torchmetrics>=0.8.0,<0.9.0',
 'transformers>=4.18.0,<5.0.0']

setup_kwargs = {
    'name': 'pytorch-ie',
    'version': '0.18.0',
    'description': 'State-of-the-art Information Extraction in PyTorch',
    'long_description': '# PyTorch-IE: State-of-the-art Information Extraction in PyTorch\n\n[![PyPI](https://img.shields.io/pypi/v/pytorch-ie.svg)][pypi status]\n[![Status](https://img.shields.io/pypi/status/pytorch-ie.svg)][pypi status]\n[![Python Version](https://img.shields.io/pypi/pyversions/pytorch-ie)][pypi status]\n[![License](https://img.shields.io/pypi/l/pytorch-ie)][license]\n\n[![Read the documentation at https://pytorch-ie.readthedocs.io/](https://img.shields.io/readthedocs/pytorch-ie/latest.svg?label=Read%20the%20Docs)][read the docs]\n[![Tests](https://github.com/christophalt/pytorch-ie/workflows/Tests/badge.svg)][tests]\n[![Codecov](https://codecov.io/gh/christophalt/pytorch-ie/branch/main/graph/badge.svg)][codecov]\n\n[![pre-commit](https://img.shields.io/badge/pre--commit-enabled-brightgreen?logo=pre-commit&logoColor=white)][pre-commit]\n[![Black](https://img.shields.io/badge/code%20style-black-000000.svg)][black]\n\n[pypi status]: https://pypi.org/project/pytorch-ie/\n[read the docs]: https://pytorch-ie.readthedocs.io/\n[tests]: https://github.com/christophalt/pytorch-ie/actions?workflow=Tests\n[codecov]: https://app.codecov.io/gh/christophalt/pytorch-ie\n[pre-commit]: https://github.com/pre-commit/pre-commit\n[black]: https://github.com/psf/black\n\n## 🤯 What\'s this about?\n\nThis is an experimental framework that aims to combine the lessons learned from five years of information extraction research.\n\n-   **Focus on the core task:** The main goal is to develop information extraction methods not dataset loading and evaluation logic. We use external well-maintained libraries for non-core functionality. PyTorch-Lightning for training and logging, Huggingface datasets for dataset reading, and Huggingface evaluate for evaluation (coming soon).\n-   **Sharing is caring:** Being able to quickly and easily share models is key to promote your work and facilitate further research. All models developed in PyTorch-IE can be easily shared via the Huggingface model hub. This further allows to quickly build demos based on Huggingface spaces, gradio or streamlit.\n-   **Unified document format:** A unified document format allows for quick experimentation on any dataset or task.\n-   **Beyond sentence level:** Most information extraction frameworks assume text inputs at a sentence granularity. We do not make any assumption on the granularity but generally aim for document-level information extraction.\n-   **Beyond unstructured text:** Unstructured text is only one possible area for information extraction. We developed the framework to also support information extraction from semi-structured text (e.g. HTML), two-dimensional text (e.g. OCR\'d images), and images.\n-   **Character-level annotation and evaluation:** Many information extraction frameworks annotate and evaluate on a token level. We believe that annotation and evaluation should be done on a character level as this also considers the suitability of the tokenizer for the task.\n-   **Make no assumptions on the structure of models:** The last years have seen many different and creative approaches to information extraction and a framework that imposes a structure on those will most certainly be to limiting. With PyTorch-iE you have full control over how a document is prepared for a model and how the model is structured. The logic is self-contained and thus can be easily shared and inspected by others. The only assumption we make is that the input is a document and the output are targets (training) or annotations (inference).\n\n## 🔭 Demos\n\n| Task                                                       | Link                                                                                                                                                                  |\n| ---------------------------------------------------------- | --------------------------------------------------------------------------------------------------------------------------------------------------------------------- |\n| Named Entity Recognition (Span-based)                      | [![Hugging Face Spaces](https://img.shields.io/badge/%F0%9F%A4%97%20Hugging%20Face-Spaces-blue)](https://huggingface.co/spaces/pie/NER)                               |\n| Joint Named Entity Recognition and Relation Classification | [![Hugging Face Spaces](https://img.shields.io/badge/%F0%9F%A4%97%20Hugging%20Face-Spaces-blue)](https://huggingface.co/spaces/pie/Joint-NER-and-Relation-Extraction) |\n\n## 🚀️ Quickstart\n\n```console\n$ pip install pytorch-ie\n```\n\n**For even faster prototyping with pre-defined, but fully configurable training pipelines and much more useful tooling, have a look into the [PyTorch-IE-Hydra-Template](https://github.com/ChristophAlt/pytorch-ie-hydra-template).**\n\n## 🥧 Concepts & Architecture\n\nPyTorch-IE builds on three core concepts: the **📃 Document**, the **🔤 ⇔ 🔢 Taskmodule**, and the **🧮 Model**. In a\nnutshell, the Document says how your data is structured, the Model defines your trainable logic and the Taskmodule\nconverts from one end to the other. All three concepts are represented as abstract classes that should be used to\nderive use-case specific versions. In the following, they are explained in detail.\n\n<details>\n<summary>\n\n### 📃 Document\n\n</summary>\n\nThe `Document` class is a special `dataclass` that defines the document model. Derivations can contain several\nelements:\n\n-   **Data fields** like strings to represent one or multiple texts or arrays for image data. These elements can be\n    arbitrary python objects.\n-   **Annotation fields** like labeled spans for entities or labeled tuples of spans for relations. These elements have\n    to be of a certain container type `AnnotationList` that is dynamically typed with the actual annotation type, e.g.\n    `entities: AnnotationList[LabeledSpan]`. Furthermore, annotation elements define one or multiple annotation `targets`.\n    An annotation target is either a data element or another annotation container. Internally, targets are used to construct the\n    annotation graph, i.e. data elements and annotation containers are the nodes and targets define the edges. The\n    annotation graph defines the (de-)serialization order and what is accessible from within an annotation. To\n    facilitate the setup of annotation containers, there is the `annotation_field()` method.\n-   **Other fields** to save metadata, ids, etc. They are not constrained in any way, but can not be accessed from within\n    annotations.\n\n<details>\n\n<summary>\n\n#### Example Document Model\n\n</summary>\n\n```python\nfrom typing import Optional\nfrom pytorch_ie.core import Document, AnnotationList, annotation_field\nfrom pytorch_ie.annotations import LabeledSpan, BinaryRelation, Label\n\nclass MyDocument(Document):\n    # data fields (any field that is targeted by an annotation fields)\n    text: str\n    # annotation fields\n    entities: AnnotationList[LabeledSpan] = annotation_field(target="text")\n    relations: AnnotationList[BinaryRelation] = annotation_field(target="entities")\n    label: AnnotationList[Label] = annotation_field()\n    # other fields\n    doc_id: Optional[str] = None\n```\n\nNote that the `label` is a special annotation field that does not define a target because it belongs to the whole document.\nYou can also have more complex constructs, like annotation fields that target multiple other fields by using\n`annotation_field(targets)` or `annotation_field(named_targets)`. The latter is useful if you want to access the\ntargets by name from within the annotation, see below for an example.\n\n</details>\n\n#### Annotations\n\nThere are several predefined **annotation types** in `pytorch_ie.annotations`, however, feel free to define your own.\nAnnotations have to be dataclasses that subclass `pytorch_ie.core.Annotation`. They also need to be hashable and\nimmutable. The following is a simple example:\n\n```python\n@dataclass(eq=True, frozen=True)\nclass SimpleLabeledSpan(Annotation):\n    start: int\n    end: int\n    label: str\n```\n\n<details>\n<summary>\n\n##### Accessing Target Content\n\n</summary>\n\nWe can expand the above example a little to have a nice string representation:\n\n```python\n@dataclass(eq=True, frozen=True)\nclass LabeledSpan(Annotation):\n    start: int\n    end: int\n    label: str\n\n    def __str__(self) -> str:\n        if self.targets is None:\n            return ""\n        return str(self.target[self.start : self.end])\n```\n\nThe content of `self.target` is lazily assigned as soon as the annotation is added to a document.\n\nNote that this now expects a single `collections.abc.Sequence` as `target`, e.g.:\n\n```python\nmy_spans: AnnotationList[Span] = annotation_field(target="<NAME_OF_THE_SEQUENCE_FIELD>")\n```\n\nIf we have multiple targets, we need to define target names to access them. For this, we need to set the special\nfield `TARGET_NAMES`:\n\n```python\n@dataclass(eq=True, frozen=True)\nclass Alignment(Annotation):\n    TARGET_NAMES = ("text1", "text2")\n    start1: int\n    end1: int\n    start2: int\n    end2: int\n\n    def __str__(self) -> str:\n        if self.targets is None:\n            return ""\n        # we can access the `named_targets` which has the keys defined in `TARGET_NAMES`\n        span1 = self.named_targets["text1"][self.start1 : self.end1]\n        span2 = self.named_targets["text2"][self.start2 : self.end2]\n        return f\'span1="{span1}" is aligned with span2="{span2}"\'\n```\n\nThis requires to define the annotation container as follows:\n\n```python\nclass MyDocumentWithAlignment(Document):\n    text_a: str\n    text_b: str\n    # `named_targets` defines the mapping from `TARGET_NAMES` to data fields\n    my_alignments: AnnotationList[Alignment] = annotation_field(named_targets={"text1": "text_a", "text2": "text_b"})\n```\n\nNote that `text1` and `text2` can also target the same field.\n\n</details>\n<details>\n<summary>\n\n##### (De-)Serialization of Annotations\n\n</summary>\n\nAs usual for dataclasses, annotations can be converted to json like objects with `.asdict()`. However, they can be\nalso created with `MyAnnotation.fromdict(dct, annotation_store)`. Both methods are required because documents and\ntheir annotations are created on the fly when working with PIE datasets (see below).\n\n</details>\n</details>\n<details>\n<summary>\n\n### 🔤 ⇔ 🔢 Taskmodule\n\n</summary>\n\nThe taskmodule is responsible for converting documents to model inputs and back. For that purpose, it requires the\nuser to implement the following methods:\n\n-   `encode_input`: Taking one document, create one or multiple `TaskEncoding`s. A `TaskEncoding` represents an\n    example that will be passed to the model later on. It is a container holding `inputs`, optional `targets`, the\n    original `document`, and `metadata`. Note that `encode_input` should not assign a value to `targets`.\n-   `encode_target`: This gets a single `TaskEncoding` and should produce a target encoding that will be assigned\n    to `targets` later on. As such, it is called only during training / evaluation, but not for inference. Note that,\n    this is allowed to return None. In this case, the respective `TaskEncoding` will not be passed to the model at all.\n-   `collate`: Taking a batch of `TaskEncoding`s, this should produce a batch input for the model. Note that this has to\n    work with available targets (training and evaluation) and without them (inference).\n-   `unbatch_output`: This gets a batch output from the model and should rearrange that into a sequence of `TaskOutput`s.\n    In that means it can be understood as the opposite to `collate`. The number of `TaskOutput`s should match the\n    number of `TaskEncoding`s that got into the batch because we align them later on for easy creation of new annotations.\n-   `create_annotations_from_output`: This gets a single `TaskEncoding` with its corresponding `TaskOutput` and\n    should yield tuples each consisting of an annotation field name and an annotation. The annotations will be added\n    as predictions to the annotation field with the respective name.\n-   `prepare` (OPTIONAL): This will get the train dataset, i.e. a Sequence or Iterable of Documents, and can be used\n    to calculate additional parameters like the list of all available labels, etc.\n\nYou can find some predefined taskmodules for _text-_ and _token classification_, _text classification based relation\nextraction_, _joint entity and relation classification_ and other use cases in the package\n[`pytorch_ie.taskmodules`](src/pytorch_ie/taskmodules). Especially, have a look at the\n[SimpleTransformerTextClassificationTaskModule](src/pytorch_ie/taskmodules/simple_transformer_text_classification.py)\nthat is well documented and should provide a good starting point to implement your own one.\n\n</details>\n<details>\n<summary>\n\n### 🧮 Model\n\n</summary>\n\nPyTorch-IE models are meant to do the heavy lifting training and inference. They are\n[Pytorch-Lightning modules](https://pytorch-lightning.readthedocs.io/en/stable/common/lightning_module.html),\nenhanced with some functionality to ease persisting them, see [Reusability and Sharing](#reusability-and-sharing).\n\nYou can find some predefined models for transformer based _text-_ and _token classification_, _sequence generation_,\nand other use cases in the package [`pytorch_ie.models`](src/pytorch_ie/models).\n\n</details>\n\n### Reusability and Sharing\n\nTaskmodules and Models provide some functionality to ease reusability and reproducibility. Especially, they provide\nthe methods `save_pretrained()` and `from_pretrained()` that can be used to save their specification, i.e. their\n**config**, and available model wights to disc and exactly re-create them again from that data.\n\n<details>\n<summary>\n\n#### Huggingface Hub and Extended Configs\n\n</summary>\n\nThese methods come along\nwith integration to the [Huggingface Hub](https://huggingface.co/docs/hub/index). By passing `push_to_hub=True` to\n`save_pretrained()`, the taskmodule / model is directly pushed to the Hub and can be loaded again with the respective\nidentifier (see the [Examples](examples) for how to do so). However, to work properly, each taskmodule / model has to\ncorrectly implement the `_config()` getter method. Per default, it returns all parameters passed to the `__init__`\nmethod if this calls `save_hyperparameters()` which is very recommended. But you may have created some further\nparameters that should be persisted, for instance a label-to-id mapping. In this case, `_config()` should be\noverwritten to take this into account:\n\n```python\ndef _config(self) -> Dict[str, Any]:\n    # add the label-to-id mapping to the config\n    config = super()._config()\n    config["label_to_id"] = self.label_to_id\n    return config\n```\n\nFurthermore, you can use the property `is_from_pretrained` to know if the taskmodule / model is just loaded or created\nfrom scratch. This may be useful, for instance, to avoid downloading a model from Huggingface Transformers when you\nin fact want to load your own trained model from disc via `from_pretrained`:\n\n```python\nfrom transformers import AutoConfig, AutoModel\n\nhf_config = AutoConfig.from_pretrained(model_name_or_path)\n# If this is already trained, just create an empty transformer model. The weights are loaded afterwards\n# via the pytorch_ie.Model.from_pretrained() logic.\nif self.is_from_pretrained:\n    self.model = AutoModel.from_config(config=hf_config)\n# Otherwise, download the whole model from the Huggingface Hub.\nelse:\n    self.model = AutoModel.from_pretrained(model_name_or_path, config=hf_config)\n```\n\n</details>\n\nIn short, each taskmodule / model implementation should:\n\n-   call `save_hyperparameters()` in `__init__` to save all constructor arguments,\n-   pass remaining `__init__` kwargs (keyword arguments) to its super to not break some other helpful functionality\n    (e.g. `is_from_pretrained`), and\n-   overwrite `_config()` if additional parameters are calculated, e.g. from the dataset.\n\n## ⚡️ Examples: Prediction\n\n**The following examples work out of the box. No further setup like manually downloading a model is needed!**\n\n**Note:** Setting `num_workers=0` in the pipeline is only necessary when running an example in an\ninteractive python session. The reason is that multiprocessing doesn\'t play well with the interactive python\ninterpreter, see [here](https://docs.python.org/3/library/multiprocessing.html#using-a-pool-of-workers)\nfor details.\n\n### Span-classification-based Named Entity Recognition\n\n```python\nfrom dataclasses import dataclass\n\nfrom pytorch_ie.annotations import LabeledSpan\nfrom pytorch_ie.auto import AutoPipeline\nfrom pytorch_ie.core import AnnotationList, annotation_field\nfrom pytorch_ie.documents import TextDocument\n\n@dataclass\nclass ExampleDocument(TextDocument):\n    entities: AnnotationList[LabeledSpan] = annotation_field(target="text")\n\ndocument = ExampleDocument(\n    "“Making a super tasty alt-chicken wing is only half of it,” said Po Bronson, general partner at SOSV and managing director of IndieBio."\n)\n\n# see below for the long version\nner_pipeline = AutoPipeline.from_pretrained("pie/example-ner-spanclf-conll03", device=-1, num_workers=0)\n\nner_pipeline(document)\n\nfor entity in document.entities.predictions:\n    print(f"{entity} -> {entity.label}")\n\n# Result:\n# IndieBio -> ORG\n# Po Bronson -> PER\n# SOSV -> ORG\n```\n\n<details>\n<summary>\nTo create the same pipeline as above without `AutoPipeline`\n</summary>\n\n```python\nfrom pytorch_ie.auto import AutoTaskModule, AutoModel\nfrom pytorch_ie.pipeline import Pipeline\n\nmodel_name_or_path = "pie/example-ner-spanclf-conll03"\nner_taskmodule = AutoTaskModule.from_pretrained(model_name_or_path)\nner_model = AutoModel.from_pretrained(model_name_or_path)\nner_pipeline = Pipeline(model=ner_model, taskmodule=ner_taskmodule, device=-1, num_workers=0)\n```\n\n</details>\n\n<details>\n<summary>\nOr, without `Auto` classes at all\n</summary>\n\n```python\nfrom pytorch_ie.pipeline import Pipeline\nfrom pytorch_ie.models import TransformerSpanClassificationModel\nfrom pytorch_ie.taskmodules import TransformerSpanClassificationTaskModule\n\nmodel_name_or_path = "pie/example-ner-spanclf-conll03"\nner_taskmodule = TransformerSpanClassificationTaskModule.from_pretrained(model_name_or_path)\nner_model = TransformerSpanClassificationModel.from_pretrained(model_name_or_path)\nner_pipeline = Pipeline(model=ner_model, taskmodule=ner_taskmodule, device=-1, num_workers=0)\n```\n\n</details>\n<details>\n<summary>\n\n### Text-classification-based Relation Extraction\n\n</summary>\n\n```python\nfrom dataclasses import dataclass\n\nfrom pytorch_ie.annotations import BinaryRelation, LabeledSpan\nfrom pytorch_ie.auto import AutoPipeline\nfrom pytorch_ie.core import AnnotationList, annotation_field\nfrom pytorch_ie.documents import TextDocument\n\n\n@dataclass\nclass ExampleDocument(TextDocument):\n    entities: AnnotationList[LabeledSpan] = annotation_field(target="text")\n    relations: AnnotationList[BinaryRelation] = annotation_field(target="entities")\n\ndocument = ExampleDocument(\n    "“Making a super tasty alt-chicken wing is only half of it,” said Po Bronson, general partner at SOSV and managing director of IndieBio."\n)\n\nre_pipeline = AutoPipeline.from_pretrained("pie/example-re-textclf-tacred", device=-1, num_workers=0)\n\nfor start, end, label in [(65, 75, "PER"), (96, 100, "ORG"), (126, 134, "ORG")]:\n    document.entities.append(LabeledSpan(start=start, end=end, label=label))\n\nre_pipeline(document, batch_size=2)\n\nfor relation in document.relations.predictions:\n    print(f"({relation.head} -> {relation.tail}) -> {relation.label}")\n\n# Result:\n# (Po Bronson -> SOSV) -> per:employee_of\n# (Po Bronson -> IndieBio) -> per:employee_of\n# (SOSV -> Po Bronson) -> org:top_members/employees\n# (IndieBio -> Po Bronson) -> org:top_members/employees\n```\n\n</details>\n\n## ⚡️ Examples: Training\n\n<details>\n\n<summary>\n\n### Span-classification-based Named Entity Recognition\n\n</summary>\n\n```python\nimport pytorch_lightning as pl\nfrom pytorch_lightning.callbacks import ModelCheckpoint\nfrom torch.utils.data import DataLoader\n\nimport datasets\nfrom pytorch_ie.models.transformer_span_classification import TransformerSpanClassificationModel\nfrom pytorch_ie.taskmodules.transformer_span_classification import (\n    TransformerSpanClassificationTaskModule,\n)\n\npl.seed_everything(42)\n\nmodel_output_path = "./model_output/"\nmodel_name = "bert-base-cased"\nnum_epochs = 10\nbatch_size = 32\n\n# Get the PIE dataset consisting of PIE Documents that will be used for training (and evaluation).\ndataset = datasets.load_dataset(\n    path="pie/conll2003",\n)\ntrain_docs, val_docs = dataset["train"], dataset["validation"]\n\nprint("train docs: ", len(train_docs))\nprint("val docs: ", len(val_docs))\n\n# Create a PIE taskmodule.\ntask_module = TransformerSpanClassificationTaskModule(\n    tokenizer_name_or_path=model_name,\n    max_length=128,\n)\n\n# Prepare the taskmodule with the training data. This may collect available labels etc.\n# The result of this should affect the state of the taskmodule config which will be\n# persisted (and can be loaded) later on.\ntask_module.prepare(train_docs)\n\n# Persist the taskmodule. This writes the taskmodule config as a json file into the\n# model_output_path directory. The config contains all constructor parameters to\n# re-create the taskmodule at this state (via AutoTaskmodule.from_pretrained(model_output_path)).\ntask_module.save_pretrained(model_output_path)\n\n# Use the taskmodule to encode the train and dev sets. This may use the text and\n# available annotations of the documents.\ntrain_dataset = task_module.encode(train_docs, encode_target=True, as_dataset=True)\nval_dataset = task_module.encode(val_docs, encode_target=True, as_dataset=True)\n\n# Create the dataloaders. Note that the taskmodule provides the collate function!\ntrain_dataloader = DataLoader(\n    train_dataset,\n    batch_size=batch_size,\n    shuffle=True,\n    collate_fn=task_module.collate,\n)\n\nval_dataloader = DataLoader(\n    val_dataset,\n    batch_size=batch_size,\n    shuffle=False,\n    collate_fn=task_module.collate,\n)\n\n# Create the PIE model. Note that we use the number of entries in the previously\n# collected label_to_id mapping to set the number of classes to predict.\nmodel = TransformerSpanClassificationModel(\n    model_name_or_path=model_name,\n    num_classes=len(task_module.label_to_id),\n    t_total=len(train_dataloader) * num_epochs,\n    learning_rate=1e-4,\n)\n\n# Optionally, set up a model checkpoint callback. See here for further information:\n# https://pytorch-lightning.readthedocs.io/en/stable/api/pytorch_lightning.callbacks.ModelCheckpoint.html\n# checkpoint_callback = ModelCheckpoint(\n#     monitor="val/f1",\n#     dirpath=model_output_path,\n#     filename="zs-ner-{epoch:02d}-val_f1-{val/f1:.2f}",\n#     save_top_k=1,\n#     mode="max",\n#     auto_insert_metric_name=False,\n#     save_weights_only=True,\n# )\n\n# Create the pytorch-lightning trainer. See here for further information:\n# https://pytorch-lightning.readthedocs.io/en/latest/api/pytorch_lightning.trainer.trainer.Trainer.html\ntrainer = pl.Trainer(\n    fast_dev_run=False,\n    max_epochs=num_epochs,\n    gpus=0,\n    enable_checkpointing=False,\n    # callbacks=[checkpoint_callback],\n    precision=32,\n)\n# Start the training.\ntrainer.fit(model, train_dataloader, val_dataloader)\n\n# Persist the trained model. This will save the model weights and the model config that allows\n# to re-create the model at this state (via AutoModel.from_pretrained(model_output_path)).\n# model.save_pretrained(model_output_path)\n```\n\n</details>\n\n## 📚 Datasets\n\nWe parse all datasets into a common format that can be loaded directly from the model hub via Huggingface datasets. The documents are cached in an arrow table and serialized / deserialized on the fly. Any changes or preprocessing applied to the documents will be cached as well.\n\n```python\nimport datasets\n\ndataset = datasets.load_dataset("pie/conll2003")\n\nprint(dataset["train"][0])\n# >>> CoNLL2003Document(text=\'EU rejects German call to boycott British lamb .\', id=\'0\', metadata={})\n\ndataset["train"][0].entities\n# >>> AnnotationList([LabeledSpan(start=0, end=2, label=\'ORG\', score=1.0), LabeledSpan(start=11, end=17, label=\'MISC\', score=1.0), LabeledSpan(start=34, end=41, label=\'MISC\', score=1.0)])\n\nentity = dataset["train"][0].entities[1]\n\nprint(f"[{entity.start}, {entity.end}] {entity}")\n# >>> [11, 17] German\n```\n\n<details>\n<summary><b>How to create your own Pytorch-IE dataset</b></summary>\n\nPyTorch-IE datasets are built on top of Huggingface datasets. For instance, consider the\n[conll2003 from the Huggingface Hub](https://huggingface.co/datasets/conll2003) and especially their respective\n[dataset loading script](https://huggingface.co/datasets/conll2003/blob/main/conll2003.py). To create a PyTorch-IE\ndataset from that, you have to implement:\n\n1. A Document class. This will be the type of individual dataset examples.\n\n```python\n@dataclass\nclass CoNLL2003Document(TextDocument):\n    entities: AnnotationList[LabeledSpan] = annotation_field(target="text")\n```\n\nHere we derive from `TextDocument` that has a simple `text` string as base annotation target. The `CoNLL2003Document`\nadds one single annotation list called `entities` that consists of `LabeledSpan`s which reference the `text` field of\nthe document. You can add further annotation types by adding `AnnotationList` fields that may also reference (i.e.\n`target`) other annotations as you like. See [\'pytorch_ie.annotations`](src/pytorch_ie/annotations.py) for predefined\nannotation types.\n\n2. A dataset config. This is similar to\n   [creating a Huggingface dataset config](https://huggingface.co/docs/datasets/dataset_script#multiple-configurations).\n\n```python\nclass CoNLL2003Config(datasets.BuilderConfig):\n    """BuilderConfig for CoNLL2003"""\n\n    def __init__(self, **kwargs):\n        """BuilderConfig for CoNLL2003.\n        Args:\n          **kwargs: keyword arguments forwarded to super.\n        """\n        super().__init__(**kwargs)\n```\n\n3. A dataset builder class. This should inherit from\n   [`pytorch_ie.data.builder.GeneratorBasedBuilder`](src/pytorch_ie/data/builder.py) which is a wrapper around the\n   [Huggingface dataset builder class](https://huggingface.co/docs/datasets/v2.4.0/en/package_reference/builder_classes#datasets.GeneratorBasedBuilder)\n   with some utility functionality to work with PyTorch-IE `Documents`. The key elements to implement are: `DOCUMENT_TYPE`,\n   `BASE_DATASET_PATH`, and `_generate_document`.\n\n```python\nclass Conll2003(pytorch_ie.data.builder.GeneratorBasedBuilder):\n    # Specify the document type. This will be the class of individual dataset examples.\n    DOCUMENT_TYPE = CoNLL2003Document\n\n    # The Huggingface identifier that points to the base dataset. This may be any string that works\n    # as path with Huggingface `datasets.load_dataset`.\n    BASE_DATASET_PATH = "conll2003"\n\n    # The builder configs, see https://huggingface.co/docs/datasets/dataset_script for further information.\n    BUILDER_CONFIGS = [\n        CoNLL2003Config(\n            name="conll2003", version=datasets.Version("1.0.0"), description="CoNLL2003 dataset"\n        ),\n    ]\n\n    # [Optional] Define additional keyword arguments which will be passed to `_generate_document` below.\n    def _generate_document_kwargs(self, dataset):\n        return {"int_to_str": dataset.features["ner_tags"].feature.int2str}\n\n    # Define how a Pytorch-IE Document will be created from a Huggingface dataset example.\n    def _generate_document(self, example, int_to_str):\n        doc_id = example["id"]\n        tokens = example["tokens"]\n        ner_tags = [int_to_str(tag) for tag in example["ner_tags"]]\n\n        text, ner_spans = tokens_and_tags_to_text_and_labeled_spans(tokens=tokens, tags=ner_tags)\n\n        document = CoNLL2003Document(text=text, id=doc_id)\n\n        for span in sorted(ner_spans, key=lambda span: span.start):\n            document.entities.append(span)\n\n        return document\n```\n\nThe full script can be found here: [datasets/conll2003/conll2003.py](datasets/conll2003/conll2003.py). Note, that to\nload the dataset with `datasets.load_dataset`, the script has to be located in a directory with the same name (as it\nis the case for standard Huggingface dataset loading scripts).\n\n</details>\n\n<!-- github-only -->\n\n✨📚✨ [Read the full documentation](https://pytorch-ie.readthedocs.io/)\n\n## 🔧 Project Development\n\n### Setup\n\nThis project is build with [Poetry](https://python-poetry.org/). It is recommended, to install Poetry via\n[pipx](https://pypa.github.io/pipx/).\n\n0. To install `pipx`, execute the following (taken from\n   [pipx installation instructions](https://pypa.github.io/pipx/installation/)):\n\n    ```bash\n    # [IF PIP IS NOT AVAILABLE] install pip\n    python -m ensurepip --upgrade\n    # [OPTIONAL] update pip\n    python -m pip install --upgrade pip\n\n    # install pipx (requires pip 19.0 or later)\n    python3 -m pip install --user pipx\n    python3 -m pipx ensurepath\n    ```\n\n    NOTE: This installs `pipx` globally!\n\n1. Install Poetry via `pipx` (or see [Poetry installation guide](https://python-poetry.org/docs/#installation)):\n\n    ```bash\n    pipx install poetry\n    ```\n\n    NOTE: This installs `pipx` globally!\n\n2. [IF REQUIRED PYTHON VERSION IS NOT AVAILABLE] install required python:\n\n    ```bash\n    # for instance, to install python3.9 on Ubuntu:\n    sudo add-apt-repository ppa:deadsnakes/ppa\n    sudo apt update\n    sudo install python3.9\n\n    # or via conda:\n    conda create -n python3.9=python3.9 -y\n    conda activate python3.9\n    ```\n\n3. Finally, install the dependencies (including for development) for PyTorch-IE:\n\n    ```bash\n    poetry install --with dev\n    ```\n\n    NOTE: If the installation gets stuck, try if disabling experimental parallel installer helps\n    ([source](https://github.com/python-poetry/poetry/issues/3352#issuecomment-732761629)):\n    `poetry config experimental.new-installer false`\n\n### Testing and code quality checks\n\nWe use [Nox](https://nox.thea.codes/en/stable/) to execute any tests and code quality tooling in a reproducible way.\n\nTo get a list of available toolchains, call:\n\n```bash\npoetry run nox -l\n```\n\nTo run a specific command from that list, call:\n\n```bash\npoetry run nox -s <command>\n```\n\nNote: To run the nox commands in the same, reproducible setup that is specified by the lock file, we call them via\n`poetry run <nox-command>`.\n\nFor instance, to run static type checking with `mypy`, call:\n\n```bash\npoetry run nox -s mypy-3.9\n```\n\nTo run all commands that also run on GitHub CI, call:\n\n```bash\npoetry run nox\n```\n\nTo run more tests (also tests marked with `@pytest.mark.slow`, but without tests for all datasets which would take forever), call:\n\n```bash\npoetry run nox -s tests_no_local_datasets-3.9\n```\n\n### Updating Dependencies\n\nCall this to update individual packages:\n\n```bash\npoetry update <package>\n```\n\nThen, commit the modified lock file to persist the state.\n\n### Releasing\n\nSince this project is based on the [Cookiecutter template](https://cookiecutter-hypermodern-python.readthedocs.io), we can follow\n[their release steps](https://cookiecutter-hypermodern-python.readthedocs.io/en/2022.6.3.post1/guide.html#how-to-make-a-release):\n\n1. Create the release branch:\n   `git switch --create release main`\n2. Increase the version:\n   `poetry version <PATCH|MINOR|MAJOR>`,\n   e.g. `poetry version patch` for a patch release\n3. Commit the changes:\n   `git commit --message="release <NEW VERSION>" pyproject.toml`,\n   e.g. `git commit --message="release 0.13.0" pyproject.toml`\n4. Push the changes to GitHub:\n   `git push origin release`\n5. Create a PR for that `release` branch on GitHub.\n6. Wait until checks passed successfully.\n7. Integrate the PR into the main branch (use `rebase` to have a linear history). This triggers the GitHub Action that\n   creates all relevant release artefacts and also uploads them to PyPI.\n8. Cleanup: Delete the `release` branch.\n\n## 🏅 Acknowledgements\n\n-   This package is based on the [sourcery-ai/python-best-practices-cookiecutter](https://github.com/sourcery-ai/python-best-practices-cookiecutter) and [cjolowicz/cookiecutter-hypermodern-python](https://github.com/cjolowicz/cookiecutter-hypermodern-python) project templates.\n\n## 📃 Citation\n\nIf you find the framework useful please consider citing it:\n\n```bibtex\n@misc{alt2022pytorchie,\n    author={Christoph Alt, Arne Binder},\n    title = {PyTorch-IE: State-of-the-art Information Extraction in PyTorch},\n    year = {2022},\n    publisher = {GitHub},\n    journal = {GitHub repository},\n    howpublished = {\\url{https://github.com/ChristophAlt/pytorch-ie}}\n}\n```\n\n[license]: https://github.com/christophalt/pytorch-ie/blob/main/LICENSE\n',
    'author': 'Christoph Alt',
    'author_email': 'christoph.alt@posteo.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/christophalt/pytorch-ie',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



"""TODO: Add a description here."""


import json
import re

import datasets

_CITATION_CHEMPROT = """\
@article{article,
author = {Kringelum, Jens and Kjaerulff, Sonny and Brunak, Søren and Lund, Ole and Oprea, Tudor and Taboureau, Olivier},
year = {2016},
month = {02},
pages = {bav123},
title = {ChemProt-3.0: A global chemical biology diseases mapping},
volume = {2016},
journal = {Database},
doi = {10.1093/database/bav123}
}"""

# You can copy an official description
_DESCRIPTION = """\
ChemProt is a publicly available compilation of chemical-protein-disease annotation resources that enables the study
of systems pharmacology for a small molecule across multiple layers of complexity from molecular to clinical levels.
In this third version, ChemProt has been updated to more than 1.7 million compounds with 7.8 million bioactivity
measurements for 19 504 proteins.
"""

_HOMEPAGE = "http://potentia.cbs.dtu.dk/ChemProt/"

# TODO: Add the license for the dataset here if you can find it
_LICENSE = ""

# TODO: Add link to the official dataset URLs here, currently pointing to preprocessed scibert files
# The HuggingFace dataset library don't host the datasets but only point to the original files
# This can be an arbitrary nested dict/list of URLs (see below in `_split_generators` method)
_DATA_URLs = {
    "train": "https://raw.githubusercontent.com/allenai/scibert/master/data/text_classification/chemprot/train.txt",
    "dev": "https://raw.githubusercontent.com/allenai/scibert/master/data/text_classification/chemprot/dev.txt",
    "test": "https://raw.githubusercontent.com/allenai/scibert/master/data/text_classification/chemprot/test.txt",
}

_CLASS_LABELS = [
    "ACTIVATOR",
    "AGONIST",
    "AGONIST-ACTIVATOR",
    "AGONIST-INHIBITOR",
    "ANTAGONIST",
    "DOWNREGULATOR",
    "INDIRECT-DOWNREGULATOR",
    "INDIRECT-UPREGULATOR",
    "INHIBITOR",
    "PRODUCT-OF",
    "SUBSTRATE",
    "SUBSTRATE_PRODUCT-OF",
    "UPREGULATOR",
]


# TODO: Name of the dataset usually match the script name with CamelCase instead of snake_case
class ChemProt(datasets.GeneratorBasedBuilder):
    """TODO: Short description of my dataset."""

    VERSION = datasets.Version("3.0.0")  # type: ignore

    def _info(self):
        return datasets.DatasetInfo(
            # This is the description that will appear on the datasets page.
            description=_DESCRIPTION,
            # This defines the different columns of the dataset and their types
            features=datasets.Features(
                {
                    "tokens": datasets.Sequence(datasets.Value("string")),
                    "head_start": datasets.Value("int32"),
                    "head_end": datasets.Value("int32"),
                    "tail_start": datasets.Value("int32"),
                    "tail_end": datasets.Value("int32"),
                    "label": datasets.ClassLabel(names=_CLASS_LABELS),
                }
            ),
            # If there's a common (input, target) tuple from the features,
            # specify them here. They'll be used if as_supervised=True in
            # builder.as_dataset.
            supervised_keys=None,
            # Homepage of the dataset for documentation
            homepage=_HOMEPAGE,
            # License for the dataset if available
            license=_LICENSE,
            # Citation for the dataset
            citation=_CITATION_CHEMPROT,
        )

    def _split_generators(self, dl_manager):
        """Returns SplitGenerators."""
        # This method is tasked with downloading/extracting the data and defining the splits depending on the configuration
        # If several configurations are possible (listed in BUILDER_CONFIGS), the configuration selected by the user is in self.config.name

        # dl_manager is a datasets.download.DownloadManager that can be used to download and extract URLs
        # It can accept any type or nested list/dict and will give back the same structure with the url replaced with path to local files.
        # By default the archives will be extracted and a path to a cached folder where they are extracted is returned instead of the archive
        data_files = dl_manager.download_and_extract(_DATA_URLs)

        return [
            datasets.SplitGenerator(
                name=datasets.Split.TRAIN,
                gen_kwargs={"filepath": data_files.get("train")},
            ),
            datasets.SplitGenerator(
                name=datasets.Split.VALIDATION,
                gen_kwargs={"filepath": data_files.get("dev")},
            ),
            datasets.SplitGenerator(
                name=datasets.Split.TEST,
                gen_kwargs={"filepath": data_files.get("test")},
            ),
        ]

    def _generate_examples(self, filepath):
        """Yields examples."""
        with open(filepath, encoding="utf-8") as f:
            for idx, line in enumerate(f.readlines()):
                example = json.loads(line)
                raw_text = example["text"]
                label = example["label"]
                id_ = str(idx)

                # handle special case with square brackets surrounding entities in raw text
                raw_text = re.sub(r"\[\[\[", "[ [[", raw_text)
                raw_text = re.sub(r"\]\]\]", "]] ]", raw_text)
                # handle unicode remnants
                raw_text = re.sub(r"(\u2002|\xa0)", " ", raw_text)

                # TODO check whether adding whitespace before and after symbols may be too aggressive
                raw_text = re.sub(r"([.,!?()])(\S)", r"\1 \2", raw_text)
                raw_text = re.sub(r"(\S)([.,!?()])", r"\1 \2", raw_text)

                # add whitespace before start marker and after end marker
                raw_text = re.sub(r"(\S)(\[\[)", r"\1 \2", raw_text)
                raw_text = re.sub(r"(\S)(<<)", r"\1 \2", raw_text)
                raw_text = re.sub(r"(\]\])(\S)", r"\1 \2", raw_text)
                raw_text = re.sub(r"(>>)(\S)", r"\1 \2", raw_text)

                tokens = raw_text.split(" ")

                assert any(e in tokens for e in ["[[", "]]", "<<", ">>"]), (
                    f"Missing head/tail markers in " f"{example}\n Tokens: {tokens}"
                )

                # Get head/tail order before determining head/tail indices and popping markers
                head_start = tokens.index("[[")
                tail_start = tokens.index("<<")
                if head_start < tail_start:
                    tokens.pop(head_start)
                    head_end = tokens.index("]]")
                    tokens.pop(head_end)
                    tail_start = tokens.index("<<")
                    tokens.pop(tail_start)
                    tail_end = tokens.index(">>")
                    tokens.pop(tail_end)
                else:
                    tokens.pop(tail_start)
                    tail_end = tokens.index(">>")
                    tokens.pop(tail_end)
                    head_start = tokens.index("[[")
                    tokens.pop(head_start)
                    head_end = tokens.index("]]")
                    tokens.pop(head_end)

                yield id_, {
                    "tokens": tokens,
                    "head_start": head_start,
                    "head_end": head_end,
                    "tail_start": tail_start,
                    "tail_end": tail_end,
                    "label": label,
                }
